#################################################################################################
#
# Copyright (c) 2017 - 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

import os
import re
import subprocess
from time import sleep

from bfloat16 import bfloat16
from cuda import cuda, cudart
import cutlass_bindings
import numpy as np

from cutlass.backend import compiler
from cutlass.backend.gemm_operation import GemmArguments, GemmOperationUniversal
from cutlass.backend.library import (
    DataTypeSize,
    DataTypeSizeBytes,
    MathOperation,
    ShortDataTypeNames,
)
from cutlass.backend.memory_manager import get_allocated_size
from cutlass.backend.reduction_operation import ReductionArguments, ReductionOperation
from cutlass.backend.test.profiler import GpuTimer
from cutlass.backend.utils.datatypes import to_cutlass
from cutlass.backend.utils.software import SubstituteTemplate


def transpose(layout):
    if layout == cutlass_bindings.RowMajor:
        return cutlass_bindings.ColumnMajor
    elif layout == cutlass_bindings.ColumnMajor:
        return cutlass_bindings.RowMajor
    elif layout == cutlass_bindings.ColumnMajorInterleaved32:
        return cutlass_bindings.RowMajorInterleaved32
    elif layout == cutlass_bindings.RowMajorInterleaved32:
        return cutlass_bindings.ColumnMajorInterleaved32


def getTensorRef(
    tensor: np.ndarray,
    problem_size: cutlass_bindings.gemm.GemmCoord,
    operand: str,
    layout: cutlass_bindings.layout,
    batch_offset: int = 0,
):
    ptr = tensor.__array_interface__["data"][0]
    if operand == "a":
        tensor_coord = problem_size.mk()
        batch_stride = problem_size.m() * problem_size.k()
    elif operand == "b":
        tensor_coord = problem_size.kn()
        batch_stride = problem_size.k() * problem_size.n()
    elif operand in ["c", "d"]:
        tensor_coord = problem_size.mn()
        batch_stride = problem_size.m() * problem_size.n()
    else:
        raise ValueError("Unknown operand: " + operand)

    elt_size = DataTypeSizeBytes[to_cutlass(tensor.dtype)]
    ptr += batch_offset * batch_stride * elt_size

    if layout == cutlass_bindings.RowMajor:
        layout = cutlass_bindings.RowMajor.packed(tensor_coord)
        layout_tag = "RowMajor"
    elif layout == cutlass_bindings.ColumnMajor:
        layout = cutlass_bindings.ColumnMajor.packed(tensor_coord)
        layout_tag = "ColumnMajor"
    elif layout == cutlass_bindings.ColumnMajorInterleaved32:
        layout = cutlass_bindings.ColumnMajorInterleaved32.packed(tensor_coord)
        layout_tag = "ColumnMajorInterleaved32"
    elif layout == cutlass_bindings.RowMajorInterleaved32:
        layout = cutlass_bindings.RowMajorInterleaved32.packed(tensor_coord)
        layout_tag = "RowMajorInterleaved32"
    else:
        raise ValueError("unsupported layout")
    if tensor.dtype == np.float32:
        ref_name = "TensorRefF32" + layout_tag
    elif tensor.dtype == np.float64:
        ref_name = "TensorRefF64" + layout_tag
    elif tensor.dtype == np.float16:
        ref_name = "TensorRefF16" + layout_tag
    elif tensor.dtype == bfloat16:
        ref_name = "TensorRefBF16" + layout_tag
    elif tensor.dtype == np.int8:
        ref_name = "TensorRefS8" + layout_tag
    elif tensor.dtype == np.int32:
        ref_name = "TensorRefS32" + layout_tag
    else:
        raise ValueError("unsupported datatype %s" % ShortDataTypeNames[tensor.dtype])

    return getattr(cutlass_bindings, ref_name)(ptr, layout)


def getTensorView(
    tensor: np.ndarray,
    problem_size: cutlass_bindings.gemm.GemmCoord,
    operand: str,
    layout: str,
    batch_offset: int = 0,
):
    tensor_ref = getTensorRef(tensor, problem_size, operand, layout, batch_offset)

    if operand == "a":
        tensor_coord = problem_size.mk()
    elif operand == "b":
        tensor_coord = problem_size.kn()
    elif operand in ["c", "d"]:
        tensor_coord = problem_size.mn()
    else:
        raise ValueError("Unknown operand: " + operand)

    if layout == cutlass_bindings.RowMajor:
        layout_tag = "RowMajor"
    elif layout == cutlass_bindings.ColumnMajor:
        layout_tag = "ColumnMajor"
    elif layout == cutlass_bindings.ColumnMajorInterleaved32:
        layout_tag = "ColumnMajorInterleaved32"
    elif layout == cutlass_bindings.RowMajorInterleaved32:
        layout_tag = "RowMajorInterleaved32"
    else:
        raise ValueError("unsupported layout")
    if tensor.dtype == np.float32:
        ref_name = "TensorViewF32" + layout_tag
    elif tensor.dtype == np.float64:
        ref_name = "TensorViewF64" + layout_tag
    elif tensor.dtype == np.float16:
        ref_name = "TensorViewF16" + layout_tag
    elif tensor.dtype == bfloat16:
        ref_name = "TensorViewBF16" + layout_tag
    elif tensor.dtype == np.int32:
        ref_name = "TensorViewS32" + layout_tag
    elif tensor.dtype == np.int8:
        ref_name = "TensorViewS8" + layout_tag
    else:
        raise ValueError("unsupported datatype")

    return getattr(cutlass_bindings, ref_name)(tensor_ref, tensor_coord)


class GemmUniversalLauncher:
    def __init__(
        self,
        operation: "GemmOperationUniversal",
        seed: int = 2080,
        interleaved=False,
        verification=True,
        profiling=False,
        warmup_iterations=500,
        iterations=500,
        **kwargs,
    ) -> None:
        # create the reduction kernel
        self.reduction_operation: ReductionOperation = ReductionOperation(
            shape=cutlass_bindings.MatrixCoord(4, 32 * operation.C.alignment),
            C=operation.C,
            element_accumulator=operation.tile_description.math_instruction.element_accumulator,
            element_compute=operation.epilogue_functor.element_epilogue,
            epilogue_functor=operation.epilogue_functor,
            count=operation.C.alignment,
        )

        self.math_operation = operation.tile_description.math_instruction.math_operation

        #: verify the output result
        self.verification = verification
        #: profile the kernel's runtime
        self.profiling = profiling

        self.timer = GpuTimer()

        self.warmup_iterations = warmup_iterations
        self.iterations = iterations

        if "sleep" in kwargs.keys():
            self.sleep_time = kwargs["sleep"]
        else:
            self.sleep_time = 0

        #
        # Compile the operator
        #

        op_list = [operation]
        if operation.arch < 90:
            # Split K via Python is currently only supported for pre-SM90 kernels
            op_list.append(self.reduction_operation)

        compiler.add_module(op_list)

        self.operation = operation

        self.dtype_A = GemmUniversalLauncher.numpy_type(operation.A.element)
        self.dtype_B = GemmUniversalLauncher.numpy_type(operation.B.element)
        self.dtype_C = GemmUniversalLauncher.numpy_type(operation.C.element)
        self.dtype_D = GemmUniversalLauncher.numpy_type(operation.C.element)

        accumulator_size = DataTypeSize[
            operation.tile_description.math_instruction.element_accumulator
        ]
        element_size = DataTypeSize[operation.A.element]

        if element_size == 1:
            self.scope_max = 1
            self.scope_min = 0
        elif element_size <= 8:
            self.scope_max = 1
            self.scope_min = -1
        elif element_size == 16:
            self.scope_max = 4
            self.scope_min = -4
        else:
            self.scope_max = 8
            self.scope_min = -8

        #: seed
        self.seed: int = seed

        #: whether the layout is interleaved
        self.interleaved = interleaved

        #: compute type
        self.compute_type = operation.epilogue_functor.element_epilogue
        self.accumulator_type = (
            operation.tile_description.math_instruction.element_accumulator
        )

    def print_problem_size(self, p, mode, batch_count):
        if mode == cutlass_bindings.gemm.Mode.Gemm:
            mode = "Gemm"
        elif mode == cutlass_bindings.gemm.Mode.Batched:
            mode = "GemmBatched"
        elif mode == cutlass_bindings.gemm.Mode.GemmSplitKParallel:
            mode = "GemmSplitKParallel"
        problem_size = "problem: %d, %d, %d\n batch_count: %d\n mode: %s" % (
            p.m(),
            p.n(),
            p.k(),
            batch_count,
            mode,
        )
        print(problem_size)

    @staticmethod
    def numpy_type(type):
        if type == cutlass_bindings.float64:
            return np.float64
        elif type == cutlass_bindings.float32:
            return np.float32
        elif type == cutlass_bindings.float16:
            return np.float16
        elif type == cutlass_bindings.bfloat16:
            return bfloat16
        elif type == cutlass_bindings.int32:
            return np.int32
        elif type == cutlass_bindings.int8:
            return np.int8
        else:
            raise ValueError("unsupported type: %s" % ShortDataTypeNames[type])

    def uniform_init(self, size, dtype):
        if dtype in [np.float32, np.float16, bfloat16, np.float64]:
            return np.ceil(
                np.random.uniform(
                    low=self.scope_min - 0.5, high=self.scope_max - 0.5, size=size
                ).astype(dtype)
            )
        else:
            return np.random.uniform(
                low=self.scope_min - 1, high=self.scope_max + 1, size=size
            ).astype(dtype)

    def reorder_tensor_B(self, tensor_B, problem_size):
        reordered_tensor_B = np.empty_like(tensor_B)
        tensor_ref_B = getTensorRef(
            tensor_B, problem_size, "b", self.operation.B.layout
        )
        reordered_tensor_ref_B = getTensorRef(
            reordered_tensor_B, problem_size, "b", self.operation.B.layout
        )
        cutlass_bindings.gemm.host.reorder_column(
            tensor_ref_B, reordered_tensor_ref_B, problem_size
        )
        return reordered_tensor_B

    def host_reference(self, problem_size, batch_count, tensor_A, tensor_B, tensor_C, alpha, beta):
        tensor_D_ref = np.ones_like(tensor_C)
        alpha = self.numpy_type(self.compute_type)(alpha)
        beta = self.numpy_type(self.compute_type)(beta)
        init_acc = 0

        alpha = self.compute_type(alpha).value()
        beta = self.compute_type(beta).value()
        init_acc = self.accumulator_type(init_acc).value()

        for i in range(batch_count):
            if self.operation.switched:
                tensor_ref_A = getTensorRef(
                    tensor_A,
                    problem_size,
                    "a",
                    transpose(self.operation.B.layout),
                    batch_offset=i,
                )
                tensor_ref_B = getTensorRef(
                    tensor_B,
                    problem_size,
                    "b",
                    transpose(self.operation.A.layout),
                    batch_offset=i,
                )
                tensor_ref_C = getTensorRef(
                    tensor_C,
                    problem_size,
                    "c",
                    transpose(self.operation.C.layout),
                    batch_offset=i,
                )
                tensor_ref_D_ref = getTensorRef(
                    tensor_D_ref,
                    problem_size,
                    "d",
                    transpose(self.operation.C.layout),
                    batch_offset=i,
                )
            else:
                tensor_ref_A = getTensorRef(
                    tensor_A, problem_size, "a", self.operation.A.layout, batch_offset=i
                )
                tensor_ref_B = getTensorRef(
                    tensor_B, problem_size, "b", self.operation.B.layout, batch_offset=i
                )
                tensor_ref_C = getTensorRef(
                    tensor_C, problem_size, "c", self.operation.C.layout, batch_offset=i
                )
                tensor_ref_D_ref = getTensorRef(
                    tensor_D_ref,
                    problem_size,
                    "d",
                    self.operation.C.layout,
                    batch_offset=i,
                )

            if self.math_operation in [MathOperation.multiply_add_saturate]:
                cutlass_bindings.test.gemm.host.gemm_saturate(
                    problem_size,
                    alpha,
                    tensor_ref_A,
                    tensor_ref_B,
                    beta,
                    tensor_ref_C,
                    tensor_ref_D_ref,
                    init_acc,
                )
            else:
                cutlass_bindings.test.gemm.host.gemm(
                    problem_size,
                    alpha,
                    tensor_ref_A,
                    tensor_ref_B,
                    beta,
                    tensor_ref_C,
                    tensor_ref_D_ref,
                    init_acc,
                )

        return tensor_D_ref

    def equal(self, tensor_D, tensor_D_ref, problem_size, batch_count):
        for i in range(batch_count):
            tensor_view_D = getTensorView(
                tensor_D, problem_size, "d", self.operation.C.layout, batch_offset=i
            )
            tensor_view_D_ref = getTensorView(
                tensor_D_ref, problem_size, "d", self.operation.C.layout, batch_offset=i
            )

            if not cutlass_bindings.test.gemm.host.equals(
                tensor_view_D, tensor_view_D_ref
            ):
                return False

        return True

    def bytes(self, problem_size, batch_count=1, alpha=1.0, beta=0.0):
        m = problem_size.m()
        n = problem_size.n()
        k = problem_size.k()

        bytes = (
            (DataTypeSize[self.operation.A.element] * m // 8) * k
            + (DataTypeSize[self.operation.B.element] * n // 8) * k
            + (DataTypeSize[self.operation.C.element] * m // 8) * n
        )

        if beta != 0:
            bytes += (DataTypeSize[self.operation.C.element] * m // 8) * n

        bytes *= batch_count

        return bytes

    def flops(self, problem_size, batch_count=1):
        m = problem_size.m()
        n = problem_size.n()
        k = problem_size.k()

        flops_ = (m * n * k) * 2 * batch_count

        return flops_

    def run_cutlass_profiler(
        self, mode, problem_size, batch_count=1, alpha=1.0, beta=0.0
    ):
        cutlass_path = os.getenv("CUTLASS_PATH")
        assert (
            cutlass_path is not None
        ), "Environment variable 'CUTLASS_PATH' is not defined."

        values = {
            "profiler_path": cutlass_path + "/build/tools/profiler/cutlass_profiler",
            "kernel_name": self.operation.procedural_name(),
            "verification_providers": "device",
            "provider": "cutlass",
            "m": str(problem_size.m()),
            "n": str(problem_size.n()),
            "k": str(problem_size.k()),
            "split_k_slices": str(batch_count),
            "alpha": str(alpha),
            "beta": str(beta),
            "warmup": str(self.warmup_iterations),
            "profile": str(self.iterations),
        }

        cmd_template = (
            "${profiler_path} --kernels=${kernel_name} --verification-providers=${verification_providers}"
            " --providers=${provider} --m=${m} --n=${n} --k=${k}"
        )

        cmd = SubstituteTemplate(cmd_template, values)
        result = subprocess.getoutput(cmd)

        m = re.search(r"Runtime:\s+(?P<runtime>\d+.\d+)", result)
        runtime = float(m.group("runtime"))

        m = re.search(r"Bytes:\s+(?P<bytes>\d+)", result)
        bytes = int(m.group("bytes"))

        m = re.search(r"FLOPs:\s+(?P<flops>\d+)", result)
        flops = int(m.group("flops"))

        # check if the problem size matches
        assert bytes == self.bytes(problem_size, alpha, beta)
        assert flops == self.flops(problem_size)

        return runtime

    def run(self, mode, problem_size, batch_count=1, split_k_slices=1, alpha=1.0, beta=0.0):
        assert get_allocated_size() == 0, (
            "%d byte of pool memory is not released in previous run"
            % get_allocated_size()
        )

        np.random.seed(self.seed)

        # Assign an actual batch count in cases where we are not running in batched mode.
        # This is to differentiate between the number of split K slices and the batch count,
        # which are overloaded within the single `batch_count` variable.
        true_batch_count = (
            batch_count if mode == cutlass_bindings.gemm.Mode.Batched else 1
        )

        tensor_A = self.uniform_init(
            size=(problem_size.m() * problem_size.k() * true_batch_count,),
            dtype=self.dtype_A,
        )
        tensor_B = self.uniform_init(
            size=(problem_size.n() * problem_size.k() * true_batch_count,),
            dtype=self.dtype_B,
        )
        tensor_C = self.uniform_init(
            size=(problem_size.m() * problem_size.n() * true_batch_count,),
            dtype=self.dtype_C,
        )
        tensor_D = np.zeros(
            shape=(problem_size.m() * problem_size.n() * true_batch_count,),
            dtype=self.dtype_D,
        )

        #
        # Launch kernel
        #

        arguments = GemmArguments(
            operation=self.operation,
            problem_size=problem_size,
            A=tensor_A,
            B=tensor_B,
            C=tensor_C,
            D=tensor_D,
            output_op=self.operation.epilogue_type(alpha, beta),
            gemm_mode=mode,
            split_k_slices=split_k_slices,
            batch=batch_count,
        )

        if mode == cutlass_bindings.gemm.Mode.GemmSplitKParallel:
            reduction_arguments = ReductionArguments(
                self.reduction_operation,
                problem_size=[problem_size.m(), problem_size.n()],
                partitions=split_k_slices,
                workspace=arguments.ptr_D,
                destination=tensor_D,
                source=tensor_C,
                output_op=self.reduction_operation.epilogue_type(alpha, beta),
            )

        self.operation.run(arguments)

        if mode == cutlass_bindings.gemm.Mode.GemmSplitKParallel:
            self.reduction_operation.run(reduction_arguments)

        passed = True

        if self.verification:
            if mode == cutlass_bindings.gemm.Mode.GemmSplitKParallel:
                reduction_arguments.sync()
            else:
                arguments.sync()
            tensor_D_ref = self.host_reference(
                problem_size,
                true_batch_count,
                tensor_A,
                tensor_B,
                tensor_C,
                alpha,
                beta,
            )
            passed = self.equal(tensor_D, tensor_D_ref, problem_size, true_batch_count)

            try:
                assert passed
            except AssertionError:
                self.print_problem_size(problem_size, mode, batch_count)

        if self.profiling:
            sleep(self.sleep_time)
            for _ in range(self.warmup_iterations):
                self.operation.run(arguments)
                if mode == cutlass_bindings.gemm.Mode.GemmSplitKParallel:
                    self.reduction_operation.run(reduction_arguments)

            self.timer.start()
            for _ in range(self.iterations):
                self.operation.run(arguments)
                if mode == cutlass_bindings.gemm.Mode.GemmSplitKParallel:
                    self.reduction_operation.run(reduction_arguments)
            self.timer.stop_and_wait()

            runtime = self.timer.duration(self.iterations)

        # free memory and clear buffers
        del arguments
        if mode == cutlass_bindings.gemm.Mode.GemmSplitKParallel:
            del reduction_arguments

        assert get_allocated_size() == 0, (
            "%d byte of pool memory is not released after current run"
            % get_allocated_size()
        )

        if self.profiling:
            return runtime
        return passed


def test_all_gemm(operation: "GemmOperationUniversal", testcase="universal"):
    passed = True

    minimum_operand_element_size = min(
        DataTypeSize[operation.A.element], DataTypeSize[operation.B.element]
    )
    opcode_class = operation.tile_description.math_instruction.opcode_class

    if opcode_class == cutlass_bindings.OpClass.Simt:
        alignment = 1
    else:
        alignment = 128 // minimum_operand_element_size

    # int8_t gemm alignment constraints
    if opcode_class == cutlass_bindings.OpClass.Simt and operation.A.element == cutlass_bindings.int8 and operation.A.layout == cutlass_bindings.ColumnMajor:
        alignment_m = 4
    else:
        alignment_m = alignment

    if (
        opcode_class == cutlass_bindings.OpClass.Simt
        and operation.B.element == cutlass_bindings.int8
        and operation.A.layout == cutlass_bindings.RowMajor
    ):
        alignment_n = 4
    else:
        alignment_n = alignment

    if (
        opcode_class == cutlass_bindings.OpClass.Simt
        and operation.A.element == cutlass_bindings.int8
        and operation.B.element == cutlass_bindings.int8
        and (
            operation.A.layout == cutlass_bindings.RowMajor
            or operation.B.layout == cutlass_bindings.ColumnMajor
        )
    ):
        alignment_k = 4
    else:
        alignment_k = alignment

    threadblock_k = operation.tile_description.threadblock_shape[2]

    if testcase == "interleaved":
        if operation.A.layout in [
            cutlass_bindings.ColumnMajorInterleaved32,
            cutlass_bindings.RowMajorInterleaved32,
        ]:
            interleavedk = 32
        else:
            raise ValueError("Unknown layout")

    # Split K mode via Python is currently only supported pre-SM90, and when stream K is not used.
    # Stream K enables split-k functionality with mode `Gemm` and a non-unit batch count.
    supports_split_k = operation.arch < 90 and not isinstance(
        operation.swizzling_functor, cutlass_bindings.ThreadblockSwizzleStreamK
    )
    if testcase == "interleaved":
        modes = [
            cutlass_bindings.gemm.Mode.Gemm,
        ]
        problem_size_m = [interleavedk, 512 + interleavedk]
        problem_size_n = [interleavedk, 512 + interleavedk]
        problem_size_k = [
            interleavedk,
            threadblock_k * operation.tile_description.stages + interleavedk,
        ]
        problem_alpha = [1.0]
        problem_beta = [0.0]
        batch_counts = [
            1,
        ]
    elif testcase == "multistage":
        modes = [
            cutlass_bindings.gemm.Mode.Gemm,
        ]
        problem_size_m = [16, 528]
        problem_size_n = [16, 528]
        problem_size_k = [
            threadblock_k,
            threadblock_k * operation.tile_description.stages
            + operation.tile_description.math_instruction.instruction_shape[2],
        ]
        problem_alpha = [1.0]
        problem_beta = [0.0]
        batch_counts = [
            1,
        ]
    else:  # universal
        modes = [cutlass_bindings.gemm.Mode.Gemm]
        batch_counts = [1, 2, 3, 5, 7]
        if supports_split_k:
            modes.append(cutlass_bindings.gemm.Mode.GemmSplitKParallel)

        problem_size_m = [alignment_m, 512 - 3 * alignment_m]
        problem_size_n = [alignment_n, 512 - 2 * alignment_n]
        if operation.tile_description.stages is None:
            stages_for_k_calc = 7
        else:
            stages_for_k_calc = operation.tile_description.stages
        problem_size_k = [
            alignment_k,
            threadblock_k * stages_for_k_calc - alignment_k,
            threadblock_k * stages_for_k_calc * 3 - alignment_k,
        ]
        problem_alpha = [1.0]
        problem_beta = [2.0]

    testbed = GemmUniversalLauncher(operation, interleaved=(testcase == "interleaved"))

    for mode in modes:
        for m in problem_size_m:
            for n in problem_size_n:
                for k in problem_size_k:
                    for batch_count in batch_counts:
                        for alpha in problem_alpha:
                            for beta in problem_beta:
                                # skip very small K problems
                                if testcase == "universal":
                                    if k // batch_count < 2 * threadblock_k:
                                        continue

                                problem_size = cutlass_bindings.gemm.GemmCoord(m, n, k)

                                if supports_split_k:
                                    split_k_slices = batch_count
                                else:
                                    split_k_slices = 1

                                overridden_mode = mode
                                if (
                                    mode == cutlass_bindings.gemm.Mode.Gemm
                                    and batch_count > 1
                                ):
                                    overridden_mode = cutlass_bindings.gemm.Mode.Batched

                                passed = testbed.run(
                                    overridden_mode,
                                    problem_size,
                                    batch_count,
                                    split_k_slices,
                                    alpha,
                                    beta,
                                )

                                (err,) = cudart.cudaDeviceSynchronize()
                                if err != cuda.CUresult.CUDA_SUCCESS:
                                    raise RuntimeError("CUDA Error %s" % str(err))

                                if not passed:
                                    return False

    return passed



# module-wide variables
import os

from cutlass.backend.arguments import *
from cutlass.backend.c_types import *
from cutlass.backend.compiler import ArtifactManager
from cutlass.backend.conv2d_operation import *
from cutlass.backend.epilogue import *
from cutlass.backend.frontend import *
from cutlass.backend.gemm_operation import *
from cutlass.backend.library import *
from cutlass.backend.memory_manager import PoolMemoryManager
from cutlass.backend.operation import *
from cutlass.backend.parser import *
from cutlass.backend.reduction_operation import *
from cutlass.backend.tensor_ref import *
from cutlass.backend.type_hint import *
from cutlass.backend.utils import *
from cutlass.backend.utils.device import device_cc
from cutlass.backend.utils.software import (
    CheckPackages,
    SubstituteTemplate,
    device_sm_count,
    get_memory_pool,
)

compiler = ArtifactManager()



#################################################################################################
#
# Copyright (c) 2023 - 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

"""
Tests emitting a CUTLASS kernel to a PyTorch CUDA extension
"""

import random
import tempfile
import unittest

import cutlass

if cutlass.utils.datatypes.torch_available:
    import torch


def _initialize(dtype, M: int, N: int, K: int):
    """
    Utility function to initialize A, B, C, and D matrices corresponding to dimensions M, N, and K

    :param dtype: data type of tensors
    :param M: M dimension of GEMM problem
    :type M: int
    :param N: N dimension of GEMM problem
    :type N: int
    :param K: N dimension of GEMM problem
    :type K: int

    :return: initialized tensors A, B, C, and D
    :rtype: list
    """
    sizes = [(M, K), (K, N), (M, N), (M, N)]
    return [torch.randint(-3, 3, size, device='cuda').to(dtype) for size in sizes]


def _generate_problems(dtype, num):
    """
    Utility function to generate `num` GEMMs of random sizes

    :param dtype: data type of tensors
    :param num: number of GEMMs to generate
    :type num: int

    :return: lists of A, B, C, and D tensors
    :rtype: list
    """
    valid_sizes = [128, 256, 512, 1024]
    As, Bs, Cs, Ds = [], [], [], []
    for _ in range(num):
        M, N, K = [random.choice(valid_sizes) for _ in range(3)]
        A, B, C, D = _initialize(dtype, M, N, K)
        As.append(A)
        Bs.append(B)
        Cs.append(C)
        Ds.append(D)
    return As, Bs, Cs, Ds


@unittest.skipIf(not cutlass.utils.datatypes.torch_available, 'PyTorch must be available to run PyTorch extension tests')
class PyTorchExtensionTest(unittest.TestCase):

    def test_gemm(self):
        random.seed(2023)

        dtype = torch.float16
        plan = cutlass.op.Gemm(element=dtype, layout=cutlass.LayoutType.RowMajor)
        plan.activation = cutlass.epilogue.relu
        op = plan.construct()

        with tempfile.TemporaryDirectory() as tmpdir:
            mod = cutlass.emit.pytorch(op, name='gemm_mod', cc=plan.cc, sourcedir=tmpdir, jit=True)

        A, B, C, _ = _initialize(dtype, 1024, 256, 512)

        D_ref = torch.nn.functional.relu(A @ B)
        D = mod.run(A, B)
        assert torch.allclose(D, D_ref)

        D = mod.run(A, B, C)
        assert torch.allclose(D, D_ref)

        D = mod.run(A, B, C, 1.0)
        assert torch.allclose(D, D_ref)

        D = mod.run(A, B, C, 1.0, 0.0)
        assert torch.allclose(D, D_ref)

        alpha = 2.0
        beta = -1.0
        D_ref = torch.nn.functional.relu((A @ B) * alpha + (beta * C))
        D = mod.run(A, B, C, alpha, beta)
        assert torch.allclose(D, D_ref)

    def test_grouped_gemm(self):
        random.seed(2023)

        dtype = torch.float16
        plan = cutlass.op.GroupedGemm(element=dtype, layout=cutlass.LayoutType.RowMajor)
        op = plan.construct()

        with tempfile.TemporaryDirectory() as tmpdir:
            mod = cutlass.emit.pytorch(op, name='grouped_gemm_mod', cc=plan.cc, sourcedir=tmpdir, jit=True)

        As, Bs, Cs, _ = _generate_problems(dtype, 50)

        def check_all(X, Y):
            for x, y in zip(X, Y):
                assert torch.allclose(x, y)

        Ds_ref = [a @ b for a, b in zip(As, Bs)]
        Ds = mod.run(As, Bs)
        check_all(Ds, Ds_ref)

        Ds = mod.run(As, Bs, Cs)
        check_all(Ds, Ds_ref)

        Ds = mod.run(As, Bs, Cs, 1.0)
        check_all(Ds, Ds_ref)

        Ds = mod.run(As, Bs, Cs, 1.0, 0.0)
        check_all(Ds, Ds_ref)

        alpha = 2.0
        beta = -1.0
        Ds_ref = [(a @ b) * alpha + (beta * c) for a, b, c in zip(As, Bs, Cs)]
        Ds = mod.run(As, Bs, Cs, alpha, beta)
        check_all(Ds, Ds_ref)


if __name__ == '__main__':
    unittest.main()



#################################################################################################
#
# Copyright (c) 2023 - 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

"""
Low-level functionality tests for GEMM with S8 operands on SM80
"""

from functools import partial

import cutlass
from cutlass.utils.datatypes import binding_opclass, binding_type
from cutlass.backend.test.gemm_testbed import test_all_gemm
import unittest

from cutlass.backend.test.utils import LayoutCombination, get_name
from cutlass.backend.utils.device import device_cc

cc = 80

# Partial specialziation for naming tests
bound_type = binding_type(cutlass.DataType.s8)
name_fn = partial(get_name, element_a=bound_type, element_b=bound_type, arch=cc)


def add_test(cls, layouts, alignments, element_output, element_accumulator,
             threadblock_shape, warp_count, stages, opclass, swizzle=None):
    """
    Create a test-running function with the given specification and set it as a method of `cls`.

    :param cls: class to which the generated method will be added
    :type cls: type
    :param layouts: layouts of A, B, and C operands
    :type layouts: list or tuple
    :param alignments: alingments of A, B, and C operands
    :type alignments: list or tuple
    :param element_output: data type of the output element
    :type element_output: cutlass.DataType
    :param element_accumulator: data type used in accumulation
    :type element_accumulator: cutlass.DataType
    :param threadblock_shape: dimensions of threadblock tiles
    :type threadblock_shape: list or tuple
    :param warp_count: warps to be launched per threadblock dimension
    :type warp_count: list or tuple
    :param stages: number of pipeline stages to use in the kernel
    :type stages: int
    :param opclass: class of operation being performed (e.g., SIMT, Tensor Core)
    :type opclass: cutlass.OpClass
    :param swizzle: threadblock swizzling functor
    """

    cluster_shape = [1, 1, 1]

    def run(self):
        """
        Dynamically-generated function that constructs a GEMM operation and verifies it against
        multiple test cases.
        """
        element_A = cutlass.DataType.s8
        element_B = cutlass.DataType.s8
        layout_A, layout_B, layout_C = layouts
        alignment_A, alignment_B, alignment_C = alignments

        plan = cutlass.op.Gemm(element_A=element_A, element_B=element_B,
                               element_C=element_output, element_D=element_output,
                               layout_A=layout_A, layout_B=layout_B, layout_C=layout_C,
                               element_accumulator=element_accumulator,
                               kernel_cc=cc)

        plan.opclass = opclass
        if swizzle is not None:
            plan.swizzling_functor = swizzle
        td = plan.tile_descriptions()[0]
        td.threadblock_shape = threadblock_shape
        td.stages = stages
        td.warp_count = warp_count
        td.cluster_shape = cluster_shape
        op = plan.construct(tile_description=td, alignment_A=alignment_A, alignment_B=alignment_B, alignment_C=alignment_C)

        self.assertTrue(test_all_gemm(op, 'universal'))

    element_epilogue = element_accumulator
    name = name_fn(layouts, alignments, binding_type(element_output), binding_type(element_accumulator),
                   binding_type(element_epilogue), cluster_shape, threadblock_shape, stages, opclass=binding_opclass(opclass))
    setattr(cls, name, run)

    return run


@unittest.skipIf(device_cc() < cc, 'Device compute capability is insufficient for SM80 tests.')
class GemmS8Sm80(unittest.TestCase):
    """
    Wrapper class to which tests will be added dynamically in __main__
    """
    pass


@unittest.skipIf(device_cc() < cc, 'Device compute capability is insufficient for SM80 tests.')
class GemmS8Sm80StreamK(unittest.TestCase):
    """
    Wrapper class to which tests will be added dynamically in __main__
    """
    pass


# Tests using TensorOp
add_test_tensorop = partial(add_test, opclass=cutlass.OpcodeClass.TensorOp)

add_test_tensorop(GemmS8Sm80, LayoutCombination.TNN, [16, 16, 16], cutlass.DataType.s8, cutlass.DataType.s32, [256, 128, 64], [4, 2, 1], 3)
add_test_tensorop(GemmS8Sm80, LayoutCombination.TNT, [16, 16, 16], cutlass.DataType.s8, cutlass.DataType.s32, [128, 256, 64], [2, 4, 1], 3)
add_test_tensorop(GemmS8Sm80, LayoutCombination.TNN, [16, 16, 4], cutlass.DataType.s32, cutlass.DataType.s32, [64, 64, 64], [1, 1, 1], 4)

# Tests using SIMT
add_test_simt = partial(add_test, opclass=cutlass.OpcodeClass.Simt)

add_test_simt(GemmS8Sm80, LayoutCombination.NNN, [1, 1, 1], cutlass.DataType.s8, cutlass.DataType.s32, [128, 128, 8], [2, 2, 1], 2)
add_test_simt(GemmS8Sm80, LayoutCombination.TNN, [1, 1, 1], cutlass.DataType.s8, cutlass.DataType.s32, [64, 128, 8], [1, 2, 1], 2)
add_test_simt(GemmS8Sm80, LayoutCombination.NTN, [1, 1, 1], cutlass.DataType.s8, cutlass.DataType.s32, [128, 64, 8], [2, 1, 1], 2)
add_test_simt(GemmS8Sm80, LayoutCombination.TTN, [1, 1, 1], cutlass.DataType.s32, cutlass.DataType.s32, [64, 64, 8], [1, 1, 1], 2)
add_test_simt(GemmS8Sm80, LayoutCombination.NNT, [1, 1, 1], cutlass.DataType.s32, cutlass.DataType.s32, [128, 128, 8], [2, 2, 1], 2)

# Stream K tests
add_test_streamk = partial(add_test, opclass=cutlass.OpcodeClass.TensorOp, swizzle=cutlass.swizzle.ThreadblockSwizzleStreamK)
add_test_streamk(GemmS8Sm80StreamK, LayoutCombination.TNT, [16, 16, 16], cutlass.DataType.s8, cutlass.DataType.s32, [128, 256, 64], [2, 4, 1], 3)


if __name__ == '__main__':
    unittest.main()



#################################################################################################
#
# Copyright (c) 2023 - 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

import logging
import os
import sys


def _cutlass_path_from_dir() -> str:
    cutlass_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../../')
    if not os.path.isdir(cutlass_path):
        raise Exception(f'Environment variable "CUTLASS_PATH" is not defined, '
                        f'and default path of {cutlass_path} does not exist.')
    return cutlass_path


def _cuda_install_path_from_nvcc() -> str:
    import subprocess
    # Attempt to detect CUDA_INSTALL_PATH based on location of NVCC
    result = subprocess.run(['which', 'nvcc'], capture_output=True)
    if result.returncode != 0:
        raise Exception(f'Unable to find nvcc via `which` utility.')

    cuda_install_path = result.stdout.decode('utf-8').split('/bin/nvcc')[0]
    if not os.path.isdir(cuda_install_path):
        raise Exception(f'Environment variable "CUDA_INSTALL_PATH" is not defined, '
                        f'and default path of {cuda_install_path} does not exist.')

    return cuda_install_path


CUTLASS_PATH = os.getenv("CUTLASS_PATH", _cutlass_path_from_dir())
CUDA_INSTALL_PATH = os.getenv("CUDA_INSTALL_PATH", _cuda_install_path_from_nvcc())
CACHE_FILE = "compiled_cache.db"

# Add the path to the CUTLASS profiler generation/manifest scripts to PYTHONPATH
sys.path.insert(0, os.path.join(CUTLASS_PATH, "tools/library/scripts/"))

# Import types/methods from the CUTLASS utility libraries for profiler generation/emission under
from library import (
    ArchitectureNames,
    DataType,
    DataTypeSize,
    EpilogueFunctor,
    GemmKind,
    LayoutTag,
    LayoutType,
    KernelScheduleSuffixes,
    KernelScheduleType,
    KernelScheduleTag,
    MathInstruction,
    MathOperation,
    OpcodeClass,
    OperationKind,
    SharedMemPerCC,
    SwizzlingFunctor,
    TensorDescription,
    TileDescription,
)

this = sys.modules[__name__]
this.logger = logging.getLogger(__name__)

def set_log_level(level: int):
    """
    Sets the log level

    :param log_level: severity of logging level to use. See https://docs.python.org/3/library/logging.html#logging-levels for options
    :type log_level: int
    """
    this.logger.setLevel(level)

set_log_level(logging.ERROR)

from cutlass.library_defaults import OptionRegistry
from cutlass.backend.utils.device import device_cc

this.option_registry = OptionRegistry(device_cc())

this.__version__ = '3.1.0'

from cutlass.backend import get_memory_pool
from cutlass.emit.pytorch import pytorch
from cutlass.op.gemm import Gemm
from cutlass.op.gemm_grouped import GroupedGemm
from cutlass.op.op import OperationBase

get_memory_pool(init_pool_size=2 ** 30, max_pool_size=2 ** 32)



#################################################################################################
#
# Copyright (c) 2017 - 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

# test/unit/conv/device/conv2d_dgrad_implicit_gemm_f32nhwc_f32nhwc_f32nhwc_simt_f32_sm80.cu
import cutlass.backend
from cutlass.backend.conv2d_operation import *
from cutlass.backend import *
from cutlass.backend.test import *
from cutlass.backend.utils.device import device_cc
import unittest


@unittest.skipIf(device_cc() < 80, "Device compute capability is insufficient for SM80 tests.")
class Conv2dDgradImplicitGemmF32nhwcF32nhwcF32nhwcSimtF32SM80(unittest.TestCase):
    def test_SM80_Device_Conv2d_Fprop_Analytic_ImplicitGemm_f32nhwc_f32nhwc_f32nhwc_simt_f32(self):
        math_inst = MathInstruction(
            instruction_shape=[1, 1, 1],
            element_a=cutlass_bindings.float32, element_b=cutlass_bindings.float32, 
            element_accumulator=cutlass_bindings.float32, opcode_class=cutlass_bindings.OpClass.Simt,
            math_operation=MathOperation.multiply_add
        )

        A = TensorDescription(
            element=math_inst.element_a, 
            layout=cutlass_bindings.TensorNHWC,
            alignment=4)
        B = TensorDescription(
            element=math_inst.element_b, 
            layout=cutlass_bindings.TensorNHWC, 
            alignment=4)
        C = TensorDescription(
            element=cutlass_bindings.float32, 
            layout=cutlass_bindings.TensorNHWC, 
            alignment=1)

        tile_description = TileDescription(
            threadblock_shape=[128, 128, 8], stages=4, 
            warp_count=[4, 2, 1],
            math_instruction=math_inst
        )

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, cutlass_bindings.float32)

        operation = Conv2dOperation(
            conv_kind=cutlass_bindings.conv.Operator.dgrad, iterator_algorithm=cutlass_bindings.conv.IteratorAlgorithm.analytic,
            arch=80, tile_description=tile_description, A=A, B=B, C=C, 
            stride_support=StrideSupport.Unity,
            epilogue_functor=epilogue_functor,
            swizzling_functor=cutlass_bindings.IdentitySwizzle1
        )
        
        self.assertTrue(test_all_conv2d(operation))
    
    def test_SM80_Device_Conv2d_Dgrad_Optimized_ImplicitGemm_f32nhwc_f32nhwc_f32nhwc_simt_f32(self):
        math_inst = MathInstruction(
            instruction_shape=[1, 1, 1],
            element_a=cutlass_bindings.float32, element_b=cutlass_bindings.float32,
            element_accumulator=cutlass_bindings.float32, opcode_class=cutlass_bindings.OpClass.Simt,
            math_operation=MathOperation.multiply_add
        )

        A = TensorDescription(
            element=math_inst.element_a, 
            layout=cutlass_bindings.TensorNHWC,
            alignment=4)
        B = TensorDescription(
            element=math_inst.element_b, 
            layout=cutlass_bindings.TensorNHWC, 
            alignment=4)
        C = TensorDescription(
            element=cutlass_bindings.float32,
            layout=cutlass_bindings.TensorNHWC, 
            alignment=1)

        tile_description = TileDescription(
            threadblock_shape=[128, 128, 8], stages=4, 
            warp_count=[2, 4, 1],
            math_instruction=math_inst
        )

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, cutlass_bindings.float32)

        operation = Conv2dOperation(
            conv_kind=cutlass_bindings.conv.Operator.dgrad, iterator_algorithm=cutlass_bindings.conv.IteratorAlgorithm.optimized,
            arch=80, tile_description=tile_description, A=A, B=B, C=C, 
            stride_support=StrideSupport.Unity,
            epilogue_functor=epilogue_functor,
            swizzling_functor=cutlass_bindings.IdentitySwizzle1
        )
        
        self.assertTrue(test_all_conv2d(operation))



if __name__ == '__main__':
    cutlass.backend.get_memory_pool(2**26, 2**26)
    unittest.main()



#################################################################################################
#
# Copyright (c) 2017 - 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

import ctypes

import cutlass_bindings
from cutlass import (
    DataType,
    KernelScheduleType
)
from cutlass.backend.library import DataTypeSizeBytes


class GemmCoord_(ctypes.Structure):
    _fields_ = [
        ("m", ctypes.c_int),
        ("n", ctypes.c_int),
        ("k", ctypes.c_int)
    ]

    def __init__(self, gemm_coord) -> None:
        for field_name, _ in self._fields_:
            setattr(self, field_name, getattr(gemm_coord, field_name)())


class GemmCoordBatched_(ctypes.Structure):
    """
    Wrapper around a GemmCoord that also contains batch count. This is used for encoding
    batched GEMM inputs to CUTLASS 3 GEMMs.
    """

    _fields_ = [
        ("m", ctypes.c_int),
        ("n", ctypes.c_int),
        ("k", ctypes.c_int),
        ("batch_count", ctypes.c_int)
    ]

    def __init__(self, gemm_coord, batch_count) -> None:
        for field_name, _ in self._fields_[:-1]:
            setattr(self, field_name, getattr(gemm_coord, field_name)())
        setattr(self, "batch_count", batch_count)


class MatrixCoord_(ctypes.Structure):
    _fields_ = [
        ("row", ctypes.c_int),
        ("column", ctypes.c_int)
    ]


class dim3_(ctypes.Structure):
    _fields_ = [
        ("x", ctypes.c_int),
        ("y", ctypes.c_int),
        ("z", ctypes.c_int)
    ]


class StrideBatched_(ctypes.Structure):
    """
    CUTLASS 3.0 strides for operands contain one static dimension and two variable dimensions. The
    variable dimensions represent the stride along non-unit-stride dimension of the row/column major
    layout, and the batch stride. This structure encodes the two variable dimensions.
    """
    _fields_ = [
        ("major_stride", ctypes.c_int64),
        ("batch_stride", ctypes.c_int64)
    ]


dtype2ctype = {
    cutlass_bindings.float16: ctypes.c_uint16,
    cutlass_bindings.float32: ctypes.c_float,
    cutlass_bindings.float64: ctypes.c_double,
    cutlass_bindings.int32: ctypes.c_int32,
}


class GenericMainloopArguments3x_(ctypes.Structure):
    """
    Structure representing the superset of possible mainloop arguments.
    This structure should not be passed to kernels directly, but, rather,
    be used as an input to one of the more specific schedule arguments, which
    will each select those arguments relevant to the particular schedule.
    """
    _fields_ = [
        ("ptr_A", ctypes.c_void_p),
        ("stride_A", StrideBatched_),
        ("ptr_B", ctypes.c_void_p),
        ("stride_B", StrideBatched_),
    ]


def get_mainloop_arguments_3x(
    kernel_schedule: KernelScheduleType,
    element_A,
    element_B,
    alignment_A: int,
    alignment_B: int) -> ctypes.Structure:
    """
    Returns the ctypes structure to be used for the 3.x kernel's mainloop parameters.

    :param kernel_schedule: type of kernel schedule to be used in the mainloop
    :type kerel_schedule: cutlass.KernelScheduleType
    :param element_A: data type of operand A
    :param element_B: data type of operand B
    :param alignment_A: alignment of operand A
    :type alignment_A: int
    :param alignment_B: alignment of operand B
    :type alignment_B: int

    :returns: ctypes structure to be used for the 3.x kernel's mainloop parameters
    :rtype: ctypes.Structure
    """
    class _MainloopArgumentsTma(ctypes.Structure):
        _fields_ = [
            ("ptr_A", ctypes.c_void_p),
            ("stride_A", StrideBatched_),
            ("ptr_B", ctypes.c_void_p),
            ("stride_B", StrideBatched_),
        ]

        @staticmethod
        def from_generic_mainloop_args(args: GenericMainloopArguments3x_):
            return _MainloopArgumentsTma(
                args.ptr_A, args.stride_A, args.ptr_B, args.stride_B,
            )

    class _MainloopArgumentsMultistage(ctypes.Structure):
        _fields_ = [
            ("ptr_A", ctypes.c_void_p),
            ("stride_A", StrideBatched_),
            ("ptr_B", ctypes.c_void_p),
            ("stride_B", StrideBatched_),
        ]

        @staticmethod
        def from_generic_mainloop_args(args: GenericMainloopArguments3x_):
            return _MainloopArgumentsMultistage(
                args.ptr_A, args.stride_A, args.ptr_B, args.stride_B,
            )

    tma_alignment_bytes = 16
    is_tma_aligned_A = ((DataTypeSizeBytes[element_A] * alignment_A) % tma_alignment_bytes) == 0
    is_tma_aligned_B = ((DataTypeSizeBytes[element_B] * alignment_B) % tma_alignment_bytes) == 0
    is_tma_aligned = is_tma_aligned_A and is_tma_aligned_B

    if kernel_schedule == KernelScheduleType.Multistage:
        return _MainloopArgumentsMultistage
    elif kernel_schedule == KernelScheduleType.ScheduleAuto:
        if is_tma_aligned:
            return _MainloopArgumentsTma
        else:
            return _MainloopArgumentsMultistage
    else:
        if is_tma_aligned:
            return _MainloopArgumentsTma
        else:
            raise Exception(f"Specified a kernel schedule using TMA ({kernel_schedule}), but "
                            "the provided data types and alignments are not properly aligned for "
                            "using TMA.")


def get_gemm_arguments_3x(mainloop_arguments, epilogue_functor):
    _EpilogueOutputOpParams = epilogue_functor.epilogue_type

    class _EpilogueArguments(ctypes.Structure):
        _fields_ = [
            ("epilogue", _EpilogueOutputOpParams),
            ("ptr_C", ctypes.c_void_p),
            ("stride_C", StrideBatched_),
            ("ptr_D", ctypes.c_void_p),
            ("stride_D", StrideBatched_),
        ]

    class _GemmArguments(ctypes.Structure):
        _fields_ = [
            ("mode", ctypes.c_int),
            ("problem_size", GemmCoordBatched_),
            ("mainloop", mainloop_arguments),
            ("epilogue", _EpilogueArguments)
        ]

    return _GemmArguments, _EpilogueArguments, _EpilogueOutputOpParams


def get_gemm_arguments(epilogue_functor):
    _EpilogueOutputOpParams = epilogue_functor.epilogue_type

    class _GemmArguments(ctypes.Structure):
        _fields_ = [
            # Arguments from UniversalArgumentsBase
            ("mode", ctypes.c_int),
            ("problem_size", GemmCoord_),
            ("batch_count", ctypes.c_int),
            ("batch_stride_D", ctypes.c_longlong),
            # Remaining arguments
            ("epilogue", _EpilogueOutputOpParams),
            ("ptr_A", ctypes.c_void_p),
            ("ptr_B", ctypes.c_void_p),
            ("ptr_C", ctypes.c_void_p),
            ("ptr_D", ctypes.c_void_p),
            ("batch_stride_A", ctypes.c_longlong),
            ("batch_stride_B", ctypes.c_longlong),
            ("batch_stride_C", ctypes.c_longlong),
            ("stride_a", ctypes.c_longlong),
            ("stride_b", ctypes.c_longlong),
            ("stride_c", ctypes.c_longlong),
            ("stride_d", ctypes.c_longlong),
            ("lda", ctypes.c_longlong),
            ("ldb", ctypes.c_longlong),
            ("ldc", ctypes.c_longlong),
            ("ldd", ctypes.c_longlong),
            ("ptr_gather_A_indices", ctypes.c_void_p),
            ("ptr_gather_B_indices", ctypes.c_void_p),
            ("ptr_scatter_D_indices", ctypes.c_void_p)
        ]

    return _GemmArguments, _EpilogueOutputOpParams


def get_gemm_arguments_streamk(epilogue_functor):
    _EpilogueOutputOpParams = epilogue_functor.epilogue_type

    class _GemmArguments(ctypes.Structure):
        _fields_ = [
            ("mode", ctypes.c_int),
            ("problem_size", GemmCoord_),
            ("batch_count", ctypes.c_int),
            ("epilogue", _EpilogueOutputOpParams),
            ("ptr_A", ctypes.c_void_p),
            ("ptr_B", ctypes.c_void_p),
            ("ptr_C", ctypes.c_void_p),
            ("ptr_D", ctypes.c_void_p),
            ("batch_stride_A", ctypes.c_longlong),
            ("batch_stride_B", ctypes.c_longlong),
            ("batch_stride_C", ctypes.c_longlong),
            ("batch_stride_D", ctypes.c_longlong),
            ("stride_a", ctypes.c_longlong),
            ("stride_b", ctypes.c_longlong),
            ("stride_c", ctypes.c_longlong),
            ("stride_d", ctypes.c_longlong),
            ("lda", ctypes.c_longlong),
            ("ldb", ctypes.c_longlong),
            ("ldc", ctypes.c_longlong),
            ("ldd", ctypes.c_longlong),
            ("avail_sms", ctypes.c_int)
        ]

    return _GemmArguments, _EpilogueOutputOpParams


###########################################################################################
# GEMM Grouped
###########################################################################################


def get_gemm_grouped_arguments(epilogue_functor):
    _EpilogueOutputOpParams = epilogue_functor.epilogue_type

    class _GEMMGroupedArguments(ctypes.Structure):
        _fields_ = [
            ("problem_sizes", ctypes.c_void_p),
            ("problem_count", ctypes.c_int),
            ("threadblock_count", ctypes.c_int),
            ("output_op", _EpilogueOutputOpParams),
            ("ptr_A", ctypes.c_void_p),
            ("ptr_B", ctypes.c_void_p),
            ("ptr_C", ctypes.c_void_p),
            ("ptr_D", ctypes.c_void_p),
            ("lda", ctypes.c_void_p),
            ("ldb", ctypes.c_void_p),
            ("ldc", ctypes.c_void_p),
            ("ldd", ctypes.c_void_p),
            ("host_problem_sizes", ctypes.c_void_p)
        ]

    return _GEMMGroupedArguments, _EpilogueOutputOpParams


############################################################################################
# Convolution2D
############################################################################################


class Conv2DProblemSize(ctypes.Structure):
    _fields_ = [
        ("N", ctypes.c_int),
        ("H", ctypes.c_int),
        ("W", ctypes.c_int),
        ("C", ctypes.c_int),
        ("P", ctypes.c_int),
        ("Q", ctypes.c_int),
        ("K", ctypes.c_int),
        ("R", ctypes.c_int),
        ("S", ctypes.c_int),
        ("pad_h", ctypes.c_int),
        ("pad_w", ctypes.c_int),
        ("stride_h", ctypes.c_int),
        ("stride_w", ctypes.c_int),
        ("dilation_h", ctypes.c_int),
        ("dilation_w", ctypes.c_int),
        ("mode", ctypes.c_int),  # kCrossCorrelation: 0, kConvolution: 1
        ("split_k_slices", ctypes.c_int),
        ("groups", ctypes.c_int)
    ]

    def __init__(self, problem_size) -> None:
        for field_name, _ in self._fields_:
            setattr(self, field_name, getattr(problem_size, field_name))


class Layout4D(ctypes.Structure):
    _fields_ = [("stride", ctypes.c_int * 3)]

    def __init__(self, tensor_ref):
        stride = tensor_ref.stride()
        setattr(self, "stride", (stride.at(0), stride.at(1), stride.at(2)))


class TensorRef_(ctypes.Structure):
    _fields_ = [
        ("ptr", ctypes.c_void_p),
        ("layout", Layout4D)
    ]

    def __init__(self, tensor_ref):
        setattr(self, "ptr", tensor_ref.data())
        setattr(self, "layout", Layout4D(tensor_ref.layout()))


class TensorRef2D_(ctypes.Structure):
    _fields_ = [
        ("ptr", ctypes.c_void_p),
        ("stride", ctypes.c_int)
    ]


def get_conv2d_arguments(epilogue_functor):
    _EpilogueOutputOpParams = epilogue_functor.epilogue_type

    class _Conv2dArguments(ctypes.Structure):
        _fields_ = [
            ("problem_size", Conv2DProblemSize),
            ("ref_A", TensorRef_),
            ("ref_B", TensorRef_),
            ("ref_C", TensorRef_),
            ("ref_D", TensorRef_),
            ("output_op", _EpilogueOutputOpParams),
            ("split_k_mode", ctypes.c_int)
        ]

    return _Conv2dArguments, _EpilogueOutputOpParams


############################################################################################
# Reduction
############################################################################################


def get_reduction_params(epilogue_functor):
    _EpilogueOutputParams = epilogue_functor.epilogue_type

    class _ReductionParams(ctypes.Structure):
        _fields_ = [
            ("problem_size", MatrixCoord_),
            ("partitions", ctypes.c_int),
            ("partition_stride", ctypes.c_longlong),
            ("workspace", TensorRef2D_),
            ("destination", TensorRef2D_),
            ("source", TensorRef2D_),
            ("output_op", _EpilogueOutputParams),
        ]

    return _ReductionParams, _EpilogueOutputParams



#################################################################################################
#
# Copyright (c) 2023 - 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

"""
Base operation used for defining high-level CUTLASS operations (e.g., GEMM, Conv2d)
"""

from bisect import bisect_left

from cutlass import option_registry
from cutlass.backend.utils.device import device_cc
from cutlass.epilogue import get_activations
from cutlass.library_defaults import _generator_ccs
from cutlass.swizzle import get_swizzling_functors


class OperationBase:
    """
    Base operation used for defining high-level CUTLASS operations (e.g., GEMM, Conv2d)
    """

    def __init__(self, cc: int = None, kernel_cc: int = None):
        """
        :param cc: compute capability of device for which kernels should be compiled. For example, if running on H100, this should be set to 90
        :type cc: int
        :param kernel_cc: compute capability of kernels to generate. For example, if running on SM90, but desiring to use a CUTLASS 2.x-style Ampere kernel, this should be set to 80
        :type kernel_cc: int
        """
        self.cc = cc if cc is not None else device_cc()
        self.specified_kernel_cc = kernel_cc is not None
        self.current_cc = kernel_cc if kernel_cc is not None else self._find_closest_cc(self.cc)
        self.tile_description = None

        self.options = option_registry.options_for_cc(self.current_cc)

        if self.options is None:
            raise Exception(f"Invalid or unsupported compute capability: {self.current_cc}")

    def _find_closest_cc(self, cc: int) -> int:
        """
        Returns the closest CC in _generator_ccs less than or equal to `cc`

        :param cc: compute capability to query
        :type cc: int

        :returns: closest CC in _generator_ccs less than or equal to `cc`
        :rtype: int
        """
        if cc in _generator_ccs:
            return cc

        # Find closest CC lower than this CC
        idx = bisect_left(_generator_ccs, cc)
        if idx == 0:
            raise Exception(f'No valid CC to fall back to for {cc}')
        return _generator_ccs[idx-1]

    def activations(self) -> list:
        """
        Returns possible activation functions that can be used

        :return: list of activation functions that can be used
        :rtype: list
        """
        return get_activations()

    def swizzling_functors(self) -> list:
        """
        Returns possible swizzling functions that can be used

        :return: list of swizzling functions that can be used
        :rtype: list
        """
        return get_swizzling_functors()

    def _reset_options(self, cc: int):
        """
        Resets the kernel options based on cc

        :param cc: compute capability to reset to
        :type cc: int
        """
        if cc != self.current_cc:
            if cc not in _generator_ccs:
                raise Exception(f'Invalid CC for CUTLASS kernels: {cc}.')
            self.current_cc = cc
            self.options = option_registry.options_for_cc(self.current_cc)



#################################################################################################
#
# Copyright (c) 2023 - 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

"""
Classes containing valid operations for a given compute capability and data types.
"""

import logging
from cuda import __version__

# Strip any additional information from the CUDA version
_cuda_version = __version__.split("rc")[0]

# Imports from CUTLASS profiler generator and manifest scripts
import generator as prof_generator
import manifest as prof_manifest

import cutlass
from cutlass.utils.check import valid_stage_count
from cutlass.utils.datatypes import td_from_profiler_td, td_from_profiler_op, has_binding_type


_generator_ccs = [50, 60, 61, 70, 75, 80, 90]


class KernelsForDataType:
    """
    Container class for keeping track of kernels that correspond to a particular combination
    of data types for operands A, B, and accumulator
    """

    def __init__(self, datatype_comb: tuple, layout_comb: tuple):
        self.datatype_comb = datatype_comb
        self.layout_comb = layout_comb

        # Dictionary mapping from alignment (int) to a list of kernels that fit the alignment
        # constraint for the data type combination
        self.kernels_by_alignment = {}

    def add(self, operation):
        """
        Add an operation to the list of supported kernels
        """
        alignment = operation.A.alignment
        if alignment not in self.kernels_by_alignment:
            self.kernels_by_alignment[alignment] = []
        self.kernels_by_alignment[alignment].append(operation)

    @property
    def alignments(self):
        """
        Returns an unsorted list of alignments supported by this data type combination

        :return: unsorted list of alignments supported by this data type combination
        :rtype: list
        """
        return list(self.kernels_by_alignment.keys())

    @property
    def all_operations(self):
        """
        Returns a list of all operations supported by this data type combination

        :return: list of all operations supported by this data type combination
        :rtype: list
        """
        ops = []
        for _, alignment_ops in self.kernels_by_alignment.items():
            ops.extend(alignment_ops)
        return ops

    def operations(self, alignment: int):
        """
        Returns operations satisfying the alignment constraint indicated by `alignment`

        :param alignment: alignment constraint of operations to return
        :type alignment: int

        :return: list of operations
        :rtype: list
        """
        if alignment not in self.kernels_by_alignment:
            raise Exception(
                f"No operations of alignment {alignment} found for data type and layout "
                f"combination {self.datatype_comb} {self.layout_comb}"
            )
        return self.kernels_by_alignment[alignment]

    def find_alignment(self, shape: tuple, layout: cutlass.LayoutType) -> int:
        """
        Returns the most preferable alignment for a given shape and layout

        :param shape: extent of each dimension of the tensor
        :type shape: tuple
        :param layout: layout of the tensor
        :type layout: cutlass.LayoutType

        :return: maximum alignment supported by the data type combination and tensor size
        :rtype: int
        """
        # Determine the leading dimension of the shape
        if layout == cutlass.LayoutType.RowMajor:
            ld = shape[0]
        elif layout == cutlass.LayoutType.RowMajor:
            ld = shape[1]
        else:
            raise Exception(f"Unexpected or unsupported layout {layout}")

        for alignment in sorted(list(self.kernels_by_alignment.keys()), reverse=True):
            if ld % alignment == 0:
                return alignment

        # Default to alignment of 1 if no others match
        return 1

    def sort(self):
        """
        Sorts each list of kernels in `kernels_by_alignment` in descending order of threadblock shape
        """
        key = lambda op: (
            op.tile_description.threadblock_shape[0]
            * op.tile_description.threadblock_shape[1]
            * op.tile_description.threadblock_shape[2]
        )
        for alignment in self.kernels_by_alignment.keys():
            self.kernels_by_alignment[alignment].sort(key=key, reverse=True)


class ArchOptions:
    """
    Structure for keeping track of kernels available on a given compute capability

    :param target_cc: compute capability of the device on which kernels will be run
    :type target_cc: int
    :param kernel_cc: compute capability of the kernels to generate
    :type kernel_cc: int
    :param operation_kind: type of operation to register
    :type operation_kind: cutlass.OperationKind
    :param gemm_kinds: types of GEMM operations that can be included
    :type gemm_kinds: list
    :param allowed_math_operations: types of primitive math operations allowed
    :type allowed_math_operations: list
    """

    def __init__(
        self,
        target_cc: int,
        kernel_cc: int,
        operation_kind: cutlass.OperationKind,
        gemm_kinds: list,
        allowed_math_operations: list = [
            cutlass.MathOperation.multiply_add,
            cutlass.MathOperation.multiply_add_saturate,
        ]
    ):
        self.cc = kernel_cc

        # Dictionary with following structure:
        #  Key: OpcodeClass
        #  Value: Dictionary with the following structure:
        #     Key: tuple of ((DataType, DataType, DataType), (LayoutType, LayoutType, LayoutType),
        #          representing ((element_a, element_b, element_accumulator), (layout_a, layout_b))
        #     Value: KernelsForDataType
        self.operations_by_opclass = {}
        self.op_class = None
        self.allowed_math_operations = allowed_math_operations

        # Identify the method within CUTLASS generator script that generates kernel
        # descriptions for the target CC
        generate_function_name = "GenerateSM" + str(kernel_cc)
        if not hasattr(prof_generator, generate_function_name):
            cutlass.logger.warning(f"No generator found for architecture {kernel_cc}")
            return
        generate_function = getattr(prof_generator, generate_function_name)

        # Initialize a default manifest and populate it with valid kernel descriptions
        # for the target CC
        args = [
            "--kernels=all",
            f"--log-level={logging.getLevelName(cutlass.logger.level)}"
        ]
        manifest_args = prof_generator.define_parser().parse_args(args)
        manifest = prof_manifest.Manifest(manifest_args)
        generate_function(manifest, _cuda_version)

        if operation_kind not in manifest.operations:
            # No kernels generated for this architecture, this could be because the CUDA
            # toolkit is insufficient to support operations in this CC
            cutlass.logger.warning(f"No operations of type {operation_kind} found for CC {kernel_cc}")
            return

        # Iterate through the available operations for this operation kind and
        # find available opclasses and data types
        for name, op_list in manifest.operations[operation_kind].items():
            for op in op_list:
                if op.gemm_kind not in gemm_kinds:
                    continue

                mi = op.tile_description.math_instruction
                if mi.math_operation not in self.allowed_math_operations:
                    continue

                datatype_comb = (mi.element_a, mi.element_b, mi.element_accumulator)

                # Skip any data types that do not currently have conversions via cutlass_bindings
                if False in [has_binding_type(elt) for elt in datatype_comb]:
                    continue

                # Prune operations that don't fit in shared memory
                td = td_from_profiler_op(op)
                if not valid_stage_count(target_cc, td)[0]:
                    continue

                if mi.opcode_class not in self.operations_by_opclass:
                    self.operations_by_opclass[mi.opcode_class] = {}

                datatype_comb = (mi.element_a, mi.element_b, mi.element_accumulator)
                layout_comb = (op.A.layout, op.B.layout)

                # Register TF32 kernels as F32 to enable F32 -> TF32 conversion + TF32 Tensor Core operations
                if datatype_comb == (cutlass.DataType.tf32, cutlass.DataType.tf32, cutlass.DataType.f32):
                    # TF32 kernels only supported on SM80 and beyond
                    if self.cc < 80:
                        continue
                    elif self.cc == 90:
                        if (op.A.element != cutlass.DataType.f32
                            or op.B.element != cutlass.DataType.f32
                            or op.C.element != cutlass.DataType.f32):
                            continue

                    datatype_comb = (cutlass.DataType.f32, cutlass.DataType.f32, cutlass.DataType.f32)

                opclass_dict = self.operations_by_opclass[mi.opcode_class]
                key = (datatype_comb, layout_comb)
                if key not in opclass_dict:
                    opclass_dict[key] = KernelsForDataType(datatype_comb, layout_comb)
                opclass_dict[key].add(op)

        # Set the default opclass to TensorOp, if available. Otherwise default to SIMT
        if cutlass.OpcodeClass.TensorOp in self.operations_by_opclass:
            self.op_class = cutlass.OpcodeClass.TensorOp
        else:
            self.op_class = cutlass.OpcodeClass.Simt

        # The profiler's generator may generate only a limited set of combinations of operands for SIMT kernels.
        # Here, we generate additional versions via a generic TileDescription.
        if cutlass.OpcodeClass.Simt not in self.operations_by_opclass:
            self.operations_by_opclass[cutlass.OpcodeClass.Simt] = {}

        types = [
            (cutlass.DataType.s8, cutlass.DataType.s8, cutlass.DataType.s8),
            (cutlass.DataType.s8, cutlass.DataType.s8, cutlass.DataType.s32),
            (cutlass.DataType.f16, cutlass.DataType.f16, cutlass.DataType.f16),
            (cutlass.DataType.f16, cutlass.DataType.f16, cutlass.DataType.f32),
            (cutlass.DataType.f32, cutlass.DataType.f32, cutlass.DataType.f32),
            (cutlass.DataType.f64, cutlass.DataType.f64, cutlass.DataType.f64),
        ]

        layouts = [
            (cutlass.LayoutType.RowMajor, cutlass.LayoutType.RowMajor),
            (cutlass.LayoutType.RowMajor, cutlass.LayoutType.ColumnMajor),
            (cutlass.LayoutType.ColumnMajor, cutlass.LayoutType.RowMajor),
            (cutlass.LayoutType.ColumnMajor, cutlass.LayoutType.ColumnMajor),
        ]
        alignment = 1
        epilogue_functor = cutlass.EpilogueFunctor.LinearCombination
        swizzling_functor = cutlass.SwizzlingFunctor.Identity8
        for type_comb in types:
            for layout_comb in layouts:
                comb = (type_comb, layout_comb)
                if comb in self.operations_by_opclass[cutlass.OpcodeClass.Simt]:
                    continue

                A = cutlass.TensorDescription(type_comb[0], layout_comb[0], alignment)
                B = cutlass.TensorDescription(type_comb[1], layout_comb[1], alignment)
                C = cutlass.TensorDescription(type_comb[2], cutlass.LayoutType.ColumnMajor, alignment)
                math_inst = cutlass.MathInstruction(
                    [1, 1, 1],
                    type_comb[0],
                    type_comb[1],
                    type_comb[2],
                    cutlass.OpcodeClass.Simt,
                    cutlass.MathOperation.multiply_add
                )

                td = cutlass.TileDescription(
                    [128, 128, 8], 2, [4, 2, 1], math_inst, 50, 1024)

                # Prune operations that don't fit in shared memory
                if not valid_stage_count(target_cc, td_from_profiler_td(td))[0]:
                    continue

                new_operation = prof_manifest.GemmOperation(
                    cutlass.GemmKind.Universal, td.minimum_compute_capability,
                    td, A, B, C, type_comb[2], epilogue_functor, swizzling_functor)

                new_kernels = KernelsForDataType(type_comb, layout_comb)
                new_kernels.add(new_operation)
                self.operations_by_opclass[cutlass.OpcodeClass.Simt][comb] = new_kernels

        # Sort all operations
        for oc in self.operations_by_opclass.keys():
            for comb in self.operations_by_opclass[oc].keys():
                self.operations_by_opclass[oc][comb].sort()

    def opclass_supports_combination(
        self, op_class: cutlass.OpcodeClass, datatype_comb: tuple, layout_comb: tuple
    ) -> bool:
        """
        Returns whether the provided operation class supports the provided data type and layout combination

        :param op_class: operation class to consider
        :type op_class: cutlass.OpcodeClass
        :param datatype_comb: tuple of data types for (element_A, element_B, element_accumulator)
        :type datatype_comb: tuple[cutlass.DataType]
        :param layout_comb: tuple of data types for (layout_A, layout_B)
        :type layout_comb: tuple[cutlass.LayoutType]

        :return: set of operation classes that support the provided data type and layout combination
        :rtype: set
        """
        if op_class not in self.operations_by_opclass:
            raise Exception(f"Unexpected or unsupported operation class {op_class}")

        return (datatype_comb, layout_comb) in self.operations_by_opclass[op_class]

    def supporting_opclasses(
        self,
        element_a: cutlass.DataType,
        element_b: cutlass.DataType,
        element_accumulator: cutlass.DataType,
        layout_a: cutlass.LayoutType,
        layout_b: cutlass.LayoutType,
    ) -> set:
        """
        Returns a set of operation classes that support the provided data type combination

        :param element_a: data type of operand A
        :type element_a: cutlass.DataType
        :param element_b: data type of operand B
        :type element_b: cutlass.DataType
        :param element_accumulator: data type of accumulator
        :type element_accumulator: cutlass.DataType
        :param layout_a: layout of operand A
        :type layout_a: cutlass.LayoutType
        :param layout_b: layout of operand B
        :type layout_b: cutlass.LayoutType

        :return: set of operation classes that support the provided data type combination
        :rtype: set
        """
        supporting_op_classes = set()
        datatype_comb = (element_a, element_b, element_accumulator)
        layout_comb = (layout_a, layout_b)

        for op_class in self.operations_by_opclass.keys():
            if self.opclass_supports_combination(op_class, datatype_comb, layout_comb):
                supporting_op_classes.add(op_class)
        return supporting_op_classes

    def operations(
        self,
        op_class: cutlass.OpcodeClass,
        element_a: cutlass.DataType,
        element_b: cutlass.DataType,
        element_accumulator: cutlass.DataType,
        layout_a: cutlass.LayoutType,
        layout_b: cutlass.LayoutType,
    ) -> KernelsForDataType:
        """
        Returns whether the provided operation class supports the provided data type combination

        :param op_class: operation class to consider
        :type op_class: cutlass.OpcodeClass
        :param element_a: data type of operand A
        :type element_a: cutlass.DataType
        :param element_b: data type of operand B
        :type element_b: cutlass.DataType
        :param element_accumulator: data type of accumulator
        :type element_accumulator: cutlass.DataType
        :param layout_a: layout of operand A
        :type layout_a: cutlass.LayoutType
        :param layout_b: layout of operand B
        :type layout_b: cutlass.LayoutType

        :return: container of kernels by alignment supported by the provided combination of parameters
        :rtype: KernelsForDataType
        """
        datatype_comb = (element_a, element_b, element_accumulator)
        layout_comb = (layout_a, layout_b)
        if not self.opclass_supports_combination(op_class, datatype_comb, layout_comb):
            raise Exception(
                f"Data type layout combination {datatype_comb}, {layout_comb} "
                f"is not supported by opcode class {op_class} on CC {self.cc}."
            )
        return self.operations_by_opclass[op_class][(datatype_comb, layout_comb)]


class OptionRegistry:
    """
    Container of all architecture-specific options

    :param target_cc: compute capability of the device on which operations will be run
    :type target_cc: int
    """

    def __init__(self, target_cc: int):
        self.registry = {}

        gemm_kinds = [cutlass.GemmKind.Universal, cutlass.GemmKind.Universal3x]
        # Construct options for each CC
        for kernel_cc in _generator_ccs:
            self.registry[kernel_cc] = ArchOptions(target_cc, kernel_cc, cutlass.OperationKind.Gemm, gemm_kinds)

    def options_for_cc(self, cc: int) -> ArchOptions:
        return self.registry.get(cc, None)



#################################################################################################
#
# Copyright (c) 2023 - 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

"""
Utility functions for converting between frontend datatypes and CUTLASS datatypes
"""

import cutlass_bindings

import cutlass
from cutlass.backend.library import (
    DataTypeSize,
    MathInstruction,
    MathOperation,
    ShortLayoutTypeNames,
    TileDescription,
)

try:
    import numpy as np

    numpy_available = True
    _library_to_numpy_dict = {
        cutlass.DataType.f16: np.float16,
        cutlass.DataType.f32: np.float32,
        cutlass.DataType.f64: np.float64,
        cutlass.DataType.s8: np.int8,
        cutlass.DataType.s32: np.int32,
    }
except ImportError:
    numpy_available = False
    _library_to_numpy_dict = {}


def numpy_library_type(inp) -> cutlass.DataType:
    if numpy_available:
        if inp == np.float16:
            return cutlass.DataType.f16
        elif inp == np.float32:
            return cutlass.DataType.f32
        elif inp == np.float64:
            return cutlass.DataType.f64
        elif inp == np.int8:
            return cutlass.DataType.s8
        elif inp == np.int32:
            return cutlass.DataType.s32
    return None


def numpy_type(inp):
    return _library_to_numpy_dict.get(inp, None)


try:
    import cupy as cp

    cupy_available = True
    _library_to_cupy_dict = {
        cutlass.DataType.f16: cp.float16,
        cutlass.DataType.f32: cp.float32,
        cutlass.DataType.f64: cp.float64,
        cutlass.DataType.s8: cp.int8,
        cutlass.DataType.s32: cp.int32,
    }
except ImportError:
    cupy_available = False
    _library_to_cupy_dict = {}


def cupy_library_type(inp) -> cutlass.DataType:
    if cupy_available:
        if inp == cp.float16:
            return cutlass.DataType.f16
        elif inp == cp.float32:
            return cutlass.DataType.f32
        elif inp == cp.float64:
            return cutlass.DataType.f64
    return None


def cupy_type(inp):
    return _library_to_cupy_dict.get(inp, None)


try:
    import torch

    torch_available = True
    _torch_to_library_dict = {
        torch.half: cutlass.DataType.f16,
        torch.float16: cutlass.DataType.f16,
        torch.float: cutlass.DataType.f32,
        torch.float32: cutlass.DataType.f32,
        torch.double: cutlass.DataType.f64,
        torch.float64: cutlass.DataType.f64,
    }

    _library_to_torch_dict = {
        cutlass.DataType.f16: torch.half,
        cutlass.DataType.f16: torch.float16,
        cutlass.DataType.f32: torch.float,
        cutlass.DataType.f32: torch.float32,
        cutlass.DataType.f64: torch.double,
        cutlass.DataType.f64: torch.float64,
    }
except ImportError:
    torch_available = False
    _torch_to_library_dict = {}
    _library_to_torch_dict = {}


def torch_library_type(inp) -> cutlass.DataType:
    return _torch_to_library_dict.get(inp, None)


def torch_type(inp):
    return _library_to_torch_dict.get(inp, None)


try:
    import bfloat16

    bfloat16_available = True
except ImportError:
    bfloat16_available = False


def bfloat16_library_type(inp) -> cutlass.DataType:
    if bfloat16_available:
        if inp == bfloat16.bfloat16:
            return cutlass.DataType.bf16


def bfloat16_type(inp) -> bfloat16.bfloat16:
    if bfloat16_available:
        if inp == cutlass.DataType.bf16:
            return bfloat16.bfloat16


# Mapping from library data type to Python-bound CUTLASS data type
library_to_binding_dict = {
    cutlass.DataType.s8: cutlass_bindings.int8,
    cutlass.DataType.s32: cutlass_bindings.int32,
    cutlass.DataType.f16: cutlass_bindings.float16,
    cutlass.DataType.bf16: cutlass_bindings.bfloat16,
    cutlass.DataType.f32: cutlass_bindings.float32,
    cutlass.DataType.f64: cutlass_bindings.float64,
    cutlass.DataType.tf32: cutlass_bindings.tfloat32,
}

# Mapping from Python-bound CUTLASS data type to library data type
binding_to_library = {
    cutlass_bindings.int8: cutlass.DataType.s8,
    cutlass_bindings.int32: cutlass.DataType.s32,
    cutlass_bindings.float16: cutlass.DataType.f16,
    cutlass_bindings.bfloat16: cutlass.DataType.bf16,
    cutlass_bindings.float32: cutlass.DataType.f32,
    cutlass_bindings.float64: cutlass.DataType.f64,
    cutlass_bindings.tfloat32: cutlass.DataType.tf32,
}


def binding_library_type(inp):
    if inp in binding_to_library:
        return binding_to_library[inp]
    return None


def has_binding_type(inp: cutlass.DataType):
    return inp in library_to_binding_dict


def library_to_binding(inp: cutlass.DataType):
    if not has_binding_type(inp):
        raise Exception(f"No available conversion from library type {inp} to Python-bound CUTLASS type")
    return library_to_binding_dict[inp]


def library_type(inp):
    if inp in cutlass.DataTypeSize.keys():
        return inp

    for cvt_fn in [
        bfloat16_library_type,
        cupy_library_type,
        numpy_library_type,
        torch_library_type,
        binding_library_type,
    ]:
        out = cvt_fn(inp)
        if out is not None:
            return out

    raise Exception(f"No available conversion from type {inp} to a library type.")


def library_layout(layout):
    if layout in cutlass.LayoutTag.keys():
        return layout

    # Convert Python-bound CUTLASS layout to profiler library layout
    if layout == cutlass_bindings.RowMajor:
        return cutlass.LayoutType.RowMajor
    elif layout == cutlass_bindings.ColumnMajor:
        return cutlass.LayoutType.ColumnMajor
    else:
        raise Exception(f"No conversion available for layout {layout} to library layout.")


def binding_type(inp):
    if inp in DataTypeSize.keys():
        return inp

    libtype = library_type(inp)
    return library_to_binding(libtype)


def binding_layout(layout):
    if layout in ShortLayoutTypeNames.keys():
        return layout
    elif layout == cutlass.LayoutType.RowMajor:
        return cutlass_bindings.RowMajor
    elif layout == cutlass.LayoutType.ColumnMajor:
        return cutlass_bindings.ColumnMajor
    else:
        raise Exception(f"No conversion available for layout {layout} to Python-bound CUTLASS layout.")


def _tensor_from_numpy(np_tensor):
    dtype = library_type(np_tensor.dtype)
    if np_tensor.flags.c_contiguous:
        layout = cutlass.LayoutType.RowMajor
    elif np_tensor.flags.f_contiguous:
        layout = cutlass.LayoutType.ColumnMajor
    return (dtype, layout)


def _tensor_from_torch(pt_tensor):
    dtype = library_type(pt_tensor.dtype)
    return (dtype, cutlass.LayoutType.RowMajor)


def get_datatype_and_layout(tensor):
    if (numpy_available and isinstance(tensor, np.ndarray)) or (
        cupy_available and isinstance(tensor, cp.ndarray)
    ):
        return _tensor_from_numpy(tensor)
    elif torch_available and isinstance(tensor, torch.Tensor):
        return _tensor_from_torch(tensor)
    else:
        raise Exception(f"Unable to convert tensor of type {type(tensor)} to Python-bound CUTLASS datatype and layout.")


def binding_opclass(opclass: cutlass.OpcodeClass):
    if opclass == cutlass.OpcodeClass.TensorOp:
        return cutlass_bindings.OpClass.TensorOp
    elif opclass == cutlass.OpcodeClass.Simt:
        return cutlass_bindings.OpClass.Simt
    else:
        raise Exception(f"Unable to convert opcode class of type {opclass} to Python-bound CUTLASS opcode class.")


_math_operation_value_map = {x.value: x for x in MathOperation}


def backend_math_operation(math_op: cutlass.MathOperation):
    if math_op.value not in _math_operation_value_map.keys():
        raise Exception(f"Unable to convert math operation of type {math_op} to backend math operation.")
    return _math_operation_value_map[math_op.value]


def construct_backend_td(td: cutlass.TileDescription,
                         kernel_schedule: cutlass.KernelScheduleType) -> TileDescription:
    mi = td.math_instruction
    backend_mi = MathInstruction(
        mi.instruction_shape,
        binding_type(mi.element_a),
        binding_type(mi.element_b),
        binding_type(mi.element_accumulator),
        binding_opclass(mi.opcode_class),
        backend_math_operation(mi.math_operation)
    )
    return TileDescription(td.threadblock_shape, td.stages, td.warp_count,
                           backend_mi, td.cluster_shape, kernel_schedule)


def td_from_profiler_op(op) -> TileDescription:
    """
    Converts the profiler's TileDescription in ``op`` into the backend TileDescription

    :param op: profiler Operation

    :returns: backend TileDescription
    :rtype: cutlass.backend.TileDescription
    """
    schedule = op.kernel_schedule if hasattr(op, 'kernel_schedule') else None
    return construct_backend_td(op.tile_description, schedule)


def td_from_profiler_td(td: cutlass.backend.TileDescription) -> TileDescription:
    """
    Converts the profiler's TileDescription into the backend TileDescription

    :param td: profiler TileDescription
    :type td: cutlass.TileDescription

    :returns: backend TileDescription
    :rtype: cutlass.backend.TileDescription
    """
    return construct_backend_td(td, kernel_schedule=None)



#################################################################################################
#
# Copyright (c) 2017 - 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

"""
Common data types and string names for them. This file is similar to /tools/library/scripts/library.py,
but uses the Pybind-bound CUTLASS data types as many keys to the dictionary.
"""

import enum

import cutlass_bindings
from cutlass import KernelScheduleType


# The following block implements enum.auto() for Python 3.5 variants that don't include it such
# as the default 3.5.2 on Ubuntu 16.04.
#
# https://codereview.stackexchange.com/questions/177309/reimplementing-pythons-enum-auto-for-compatibility

try:
    from enum import auto as enum_auto
except ImportError:
    __cutlass_library_auto_enum = 0

    def enum_auto() -> int:
        global __cutlass_library_auto_enum
        i = __cutlass_library_auto_enum
        __cutlass_library_auto_enum += 1
        return i


ShortDataTypeNames = {
    cutlass_bindings.int32: "i",
    cutlass_bindings.float16: "h",
    cutlass_bindings.float32: "s",
    cutlass_bindings.float64: "d",
    cutlass_bindings.dtype.cf32: "c",
    cutlass_bindings.dtype.cf64: "z",
}


DataTypeNames = {
    cutlass_bindings.dtype.b1: "b1",
    cutlass_bindings.dtype.u4: "u4",
    cutlass_bindings.dtype.u8: "u8",
    cutlass_bindings.dtype.u16: "u16",
    cutlass_bindings.dtype.u32: "u32",
    cutlass_bindings.dtype.u64: "u64",
    cutlass_bindings.dtype.s4: "s4",
    cutlass_bindings.int8: "s8",
    cutlass_bindings.dtype.s16: "s16",
    cutlass_bindings.int32: "s32",
    cutlass_bindings.dtype.s64: "s64",
    cutlass_bindings.float16: "f16",
    cutlass_bindings.bfloat16: "bf16",
    cutlass_bindings.float32: "f32",
    cutlass_bindings.tfloat32: "tf32",
    cutlass_bindings.float64: "f64",
    cutlass_bindings.dtype.cf16: "cf16",
    cutlass_bindings.dtype.cbf16: "cbf16",
    cutlass_bindings.dtype.cf32: "cf32",
    cutlass_bindings.dtype.ctf32: "ctf32",
    cutlass_bindings.dtype.cf64: "cf64",
    cutlass_bindings.dtype.cu4: "cu4",
    cutlass_bindings.dtype.cu8: "cu8",
    cutlass_bindings.dtype.cu16: "cu16",
    cutlass_bindings.dtype.cu32: "cu32",
    cutlass_bindings.dtype.cu64: "cu64",
    cutlass_bindings.dtype.cs4: "cs4",
    cutlass_bindings.dtype.cs8: "cs8",
    cutlass_bindings.dtype.cs16: "cs16",
    cutlass_bindings.dtype.cs32: "cs32",
    cutlass_bindings.dtype.cs64: "cs64",
}


DataTypeTag = {
    cutlass_bindings.dtype.b1: "cutlass::uint1b_t",
    cutlass_bindings.dtype.u4: "cutlass::uint4b_t",
    cutlass_bindings.dtype.u8: "uint8_t",
    cutlass_bindings.dtype.u16: "uint16_t",
    cutlass_bindings.dtype.u32: "uint32_t",
    cutlass_bindings.dtype.u64: "uint64_t",
    cutlass_bindings.dtype.s4: "cutlass::int4b_t",
    cutlass_bindings.int8: "int8_t",
    cutlass_bindings.dtype.s16: "int16_t",
    cutlass_bindings.int32: "int32_t",
    cutlass_bindings.dtype.s64: "int64_t",
    cutlass_bindings.float16: "cutlass::half_t",
    cutlass_bindings.bfloat16: "cutlass::bfloat16_t",
    cutlass_bindings.float32: "float",
    cutlass_bindings.tfloat32: "cutlass::tfloat32_t",
    cutlass_bindings.float64: "double",
    cutlass_bindings.dtype.cf16: "cutlass::complex<cutlass::half_t>",
    cutlass_bindings.dtype.cbf16: "cutlass::complex<cutlass::bfloat16_t>",
    cutlass_bindings.dtype.cf32: "cutlass::complex<float>",
    cutlass_bindings.dtype.ctf32: "cutlass::complex<cutlass::tfloat32_t>",
    cutlass_bindings.dtype.cf64: "cutlass::complex<double>",
    cutlass_bindings.dtype.cu4: "cutlass::complex<cutlass::uint4b_t>",
    cutlass_bindings.dtype.cu8: "cutlass::complex<cutlass::uint8_t>",
    cutlass_bindings.dtype.cu16: "cutlass::complex<cutlass::uint16_t>",
    cutlass_bindings.dtype.cu32: "cutlass::complex<cutlass::uint32_t>",
    cutlass_bindings.dtype.cu64: "cutlass::complex<cutlass::uint64_t>",
    cutlass_bindings.dtype.cs4: "cutlass::complex<cutlass::int4b_t>",
    cutlass_bindings.dtype.cs8: "cutlass::complex<cutlass::int8_t>",
    cutlass_bindings.dtype.cs16: "cutlass::complex<cutlass::int16_t>",
    cutlass_bindings.dtype.cs32: "cutlass::complex<cutlass::int32_t>",
    cutlass_bindings.dtype.cs64: "cutlass::complex<cutlass::int64_t>",
}


DataTypeSize = {
    cutlass_bindings.dtype.b1: 1,
    cutlass_bindings.dtype.u4: 4,
    cutlass_bindings.dtype.u8: 8,
    cutlass_bindings.dtype.u16: 16,
    cutlass_bindings.dtype.u32: 32,
    cutlass_bindings.dtype.u64: 64,
    cutlass_bindings.dtype.s4: 4,
    cutlass_bindings.int8: 8,
    cutlass_bindings.dtype.s16: 16,
    cutlass_bindings.int32: 32,
    cutlass_bindings.dtype.s64: 64,
    cutlass_bindings.float16: 16,
    cutlass_bindings.bfloat16: 16,
    cutlass_bindings.float32: 32,
    cutlass_bindings.tfloat32: 32,
    cutlass_bindings.float64: 64,
    cutlass_bindings.dtype.cf16: 32,
    cutlass_bindings.dtype.cbf16: 32,
    cutlass_bindings.dtype.cf32: 64,
    cutlass_bindings.dtype.ctf32: 32,
    cutlass_bindings.dtype.cf64: 128,
    cutlass_bindings.dtype.cu4: 8,
    cutlass_bindings.dtype.cu8: 16,
    cutlass_bindings.dtype.cu16: 32,
    cutlass_bindings.dtype.cu32: 64,
    cutlass_bindings.dtype.cu64: 128,
    cutlass_bindings.dtype.cs4: 8,
    cutlass_bindings.dtype.cs8: 16,
    cutlass_bindings.dtype.cs16: 32,
    cutlass_bindings.dtype.cs32: 64,
    cutlass_bindings.dtype.cs64: 128,
}


class DataTypeSizeBytes:
    """
    Static class to mimic the `DataTypeSize` dictionary, but with checks for whether the
    data type key is less than a full byte or a non-integer number of bytes.
    """

    @staticmethod
    def __class_getitem__(datatype):
        """
        Returns the number of bytes in size the data type is. Raises an exception if the data type
        is either less than a full byte or a non-integer number of bytes in size.

        :param datatype: data type to query

        :return: number of bytes the data type occupies
        :rtype: int
        """
        bits = DataTypeSize[datatype]
        if bits < 8:
            raise Exception(
                "Data type {} is less than one byte in size.".format(datatype)
            )
        elif bits % 8 != 0:
            raise Exception(
                "Data type {} is not an integer number of bytes.".format(datatype)
            )
        return bits // 8


ComplexTransformTag = {
    cutlass_bindings.complex_transform.none: "cutlass::ComplexTransform::kNone",
    cutlass_bindings.complex_transform.conj: "cutlass::ComplexTransform::kConjugate",
}


RealComplexBijection = [
    (cutlass_bindings.float16, cutlass_bindings.dtype.cf16),
    (cutlass_bindings.float32, cutlass_bindings.dtype.cf32),
    (cutlass_bindings.float64, cutlass_bindings.dtype.cf64),
]


def is_complex(data_type):
    for r, c in RealComplexBijection:
        if data_type == c:
            return True
    return False


def get_complex_from_real(real_type):
    for r, c in RealComplexBijection:
        if real_type == r:
            return c
    return cutlass_bindings.dtype.invalid


def get_real_from_complex(complex_type):
    for r, c in RealComplexBijection:
        if complex_type == c:
            return r
    return cutlass_bindings.dtype.invalid


class ComplexMultiplyOp(enum.Enum):
    multiply_add = enum_auto()
    gaussian = enum_auto()


class MathOperation(enum.Enum):
    multiply_add = enum_auto()
    multiply_add_saturate = enum_auto()
    xor_popc = enum_auto()
    multiply_add_fast_bf16 = enum_auto()
    multiply_add_fast_f16 = enum_auto()
    multiply_add_fast_f32 = enum_auto()
    multiply_add_complex_fast_f32 = enum_auto()
    multiply_add_complex = enum_auto()
    multiply_add_complex_gaussian = enum_auto()


MathOperationNames = {
    MathOperation.multiply_add: "multiply_add",
    MathOperation.multiply_add_saturate: "multiply_add_saturate",
    MathOperation.xor_popc: "xor_popc",
    MathOperation.multiply_add_fast_bf16: "multiply_add_fast_bf16",
    MathOperation.multiply_add_fast_f16: "multiply_add_fast_f16",
    MathOperation.multiply_add_fast_f32: "multiply_add_fast_f32",
    MathOperation.multiply_add_complex_fast_f32: "multiply_add_complex_fast_f32",
    MathOperation.multiply_add_complex: "multiply_add_complex",
    MathOperation.multiply_add_complex_gaussian: "multiply_add_complex_gaussian",
}


MathOperationTag = {
    MathOperation.multiply_add: "cutlass::arch::OpMultiplyAdd",
    MathOperation.multiply_add_saturate: "cutlass::arch::OpMultiplyAddSaturate",
    MathOperation.xor_popc: "cutlass::arch::OpXorPopc",
    MathOperation.multiply_add_fast_bf16: "cutlass::arch::OpMultiplyAddFastBF16",
    MathOperation.multiply_add_fast_f16: "cutlass::arch::OpMultiplyAddFastF16",
    MathOperation.multiply_add_fast_f32: "cutlass::arch::OpMultiplyAddFastF32",
    MathOperation.multiply_add_complex_fast_f32: "cutlass::arch::OpMultiplyAddComplexFastF32",
    MathOperation.multiply_add_complex: "cutlass::arch::OpMultiplyAddComplex",
    MathOperation.multiply_add_complex_gaussian: "cutlass::arch::OpMultiplyAddGaussianComplex",
}


LayoutTag = {
    cutlass_bindings.ColumnMajor: "cutlass::layout::ColumnMajor",
    cutlass_bindings.RowMajor: "cutlass::layout::RowMajor",
    cutlass_bindings.layout.ColumnMajorInterleaved2: "cutlass::layout::ColumnMajorInterleaved<2>",
    cutlass_bindings.layout.RowMajorInterleaved2: "cutlass::layout::RowMajorInterleaved<2>",
    cutlass_bindings.ColumnMajorInterleaved32: "cutlass::layout::ColumnMajorInterleaved<32>",
    cutlass_bindings.RowMajorInterleaved32: "cutlass::layout::RowMajorInterleaved<32>",
    cutlass_bindings.layout.ColumnMajorInterleaved64: "cutlass::layout::ColumnMajorInterleaved<64>",
    cutlass_bindings.layout.RowMajorInterleaved64: "cutlass::layout::RowMajorInterleaved<64>",
    cutlass_bindings.TensorNHWC: "cutlass::layout::TensorNHWC",
    cutlass_bindings.layout.TensorNDHWC: "cutlass::layout::TensorNDHWC",
    cutlass_bindings.layout.TensorNCHW: "cutlass::layout::TensorNCHW",
    cutlass_bindings.layout.TensorNGHWC: "cutlass::layout::TensorNGHWC",
    cutlass_bindings.TensorNC32HW32: "cutlass::layout::TensorNCxHWx<32>",
    cutlass_bindings.TensorC32RSK32: "cutlass::layout::TensorCxRSKx<32>",
    cutlass_bindings.layout.TensorNC64HW64: "cutlass::layout::TensorNCxHWx<64>",
    cutlass_bindings.layout.TensorC64RSK64: "cutlass::layout::TensorCxRSKx<64>",
}


TransposedLayout = {
    cutlass_bindings.ColumnMajor: cutlass_bindings.RowMajor,
    cutlass_bindings.RowMajor: cutlass_bindings.ColumnMajor,
    cutlass_bindings.layout.ColumnMajorInterleaved2: cutlass_bindings.layout.RowMajorInterleaved2,
    cutlass_bindings.layout.RowMajorInterleaved2: cutlass_bindings.layout.ColumnMajorInterleaved2,
    cutlass_bindings.ColumnMajorInterleaved32: cutlass_bindings.RowMajorInterleaved32,
    cutlass_bindings.RowMajorInterleaved32: cutlass_bindings.ColumnMajorInterleaved32,
    cutlass_bindings.layout.ColumnMajorInterleaved64: cutlass_bindings.layout.RowMajorInterleaved64,
    cutlass_bindings.layout.RowMajorInterleaved64: cutlass_bindings.layout.ColumnMajorInterleaved64,
    cutlass_bindings.TensorNHWC: cutlass_bindings.TensorNHWC,
}


ShortLayoutTypeNames = {
    cutlass_bindings.ColumnMajor: "n",
    cutlass_bindings.layout.ColumnMajorInterleaved2: "n2",
    cutlass_bindings.ColumnMajorInterleaved32: "n32",
    cutlass_bindings.layout.ColumnMajorInterleaved64: "n64",
    cutlass_bindings.RowMajor: "t",
    cutlass_bindings.layout.RowMajorInterleaved2: "t2",
    cutlass_bindings.RowMajorInterleaved32: "t32",
    cutlass_bindings.layout.RowMajorInterleaved64: "t64",
    cutlass_bindings.TensorNHWC: "nhwc",
    cutlass_bindings.layout.TensorNDHWC: "ndhwc",
    cutlass_bindings.layout.TensorNCHW: "nchw",
    cutlass_bindings.layout.TensorNGHWC: "nghwc",
    cutlass_bindings.TensorNC32HW32: "nc32hw32",
    cutlass_bindings.layout.TensorNC64HW64: "nc64hw64",
    cutlass_bindings.TensorC32RSK32: "c32rsk32",
    cutlass_bindings.layout.TensorC64RSK64: "c64rsk64",
}


ShortComplexLayoutNames = {
    (cutlass_bindings.ColumnMajor, cutlass_bindings.complex_transform.none): "n",
    (cutlass_bindings.ColumnMajor, cutlass_bindings.complex_transform.conj): "c",
    (cutlass_bindings.RowMajor, cutlass_bindings.complex_transform.none): "t",
    (cutlass_bindings.RowMajor, cutlass_bindings.complex_transform.conj): "h",
}


OpcodeClassNames = {
    cutlass_bindings.OpClass.Simt: "simt",
    cutlass_bindings.OpClass.TensorOp: "tensorop",
    cutlass_bindings.OpClass.WmmaTensorOp: "wmma_tensorop",
    cutlass_bindings.OpClass.SparseTensorOp: "sptensorop",
}


OpcodeClassTag = {
    cutlass_bindings.OpClass.Simt: "cutlass::arch::OpClassSimt",
    cutlass_bindings.OpClass.TensorOp: "cutlass::arch::OpClassTensorOp",
    cutlass_bindings.OpClass.WmmaTensorOp: "cutlass::arch::OpClassWmmaTensorOp",
    cutlass_bindings.OpClass.SparseTensorOp: "cutlass::arch::OpClassSparseTensorOp",
}


class OperationKind(enum.Enum):
    Gemm = enum_auto()
    Conv2d = enum_auto()
    Conv3d = enum_auto()


OperationKindNames = {
    OperationKind.Gemm: "gemm",
    OperationKind.Conv2d: "conv2d",
    OperationKind.Conv3d: "conv3d",
}


ArchitectureNames = {
    50: "maxwell",
    60: "pascal",
    61: "pascal",
    70: "volta",
    75: "turing",
    80: "ampere",
    90: "hopper",
}


SharedMemPerCC = {
    70: 96 << 10,  # 96KB of SMEM
    72: 96 << 10,  # 96KB of SMEM
    75: 64 << 10,  # 64KB of SMEM
    80: 160 << 10,  # 164KB of SMEM - 4KB reserved for the driver
    86: 100 << 10,  # 100KB of SMEM
    87: 160 << 10,  # 164KB of SMEM - 4KB reserved for the driver
    89: 100 << 10,  # 100KB of SMEM
    90: 227 << 10,  # 228KB of SMEM - 1KB reserved for the driver
}


class GemmKind(enum.Enum):
    Gemm = enum_auto()
    Sparse = enum_auto()
    Universal = enum_auto()
    PlanarComplex = enum_auto()
    PlanarComplexArray = enum_auto()
    Grouped = enum_auto()


GemmKindNames = {
    GemmKind.Gemm: "gemm",
    GemmKind.Sparse: "spgemm",
    GemmKind.Universal: "gemm",
    GemmKind.PlanarComplex: "gemm_planar_complex",
    GemmKind.PlanarComplexArray: "gemm_planar_complex_array",
    GemmKind.Grouped: "gemm_grouped",
}


class SwizzlingFunctor(enum.Enum):
    Identity1 = enum_auto()
    Identity2 = enum_auto()
    Identity4 = enum_auto()
    Identity8 = enum_auto()
    Horizontal = enum_auto()
    BatchedIdentity1 = enum_auto()
    StridedDgradIdentity1 = enum_auto()
    StridedDgradIdentity4 = enum_auto()
    StridedDgradHorizontal = enum_auto()


SwizzlingFunctorTag = {
    cutlass_bindings.IdentitySwizzle1: "cutlass::gemm::threadblock::GemmIdentityThreadblockSwizzle<1>",
    SwizzlingFunctor.Identity2: "cutlass::gemm::threadblock::GemmIdentityThreadblockSwizzle<2>",
    SwizzlingFunctor.Identity4: "cutlass::gemm::threadblock::GemmIdentityThreadblockSwizzle<4>",
    SwizzlingFunctor.Identity8: "cutlass::gemm::threadblock::GemmIdentityThreadblockSwizzle<8>",
    SwizzlingFunctor.Horizontal: "cutlass::gemm::threadblock::GemmHorizontalThreadblockSwizzle",
    SwizzlingFunctor.BatchedIdentity1: "cutlass::gemm::threadblock::GemmBatchedIdentityThreadblockSwizzle",
    SwizzlingFunctor.StridedDgradIdentity1: "cutlass::conv::threadblock::StridedDgradIdentityThreadblockSwizzle<1>",
    SwizzlingFunctor.StridedDgradIdentity4: "cutlass::conv::threadblock::StridedDgradIdentityThreadblockSwizzle<4>",
    SwizzlingFunctor.StridedDgradHorizontal: "cutlass::conv::threadblock::StridedDgradHorizontalThreadblockSwizzle",
}


class SchedulerMode(enum.Enum):
    Device = (enum_auto(),)
    Host = enum_auto()


SchedulerModeTag = {
    SchedulerMode.Device: "cutlass::gemm::kernel::GroupScheduleMode::kDeviceOnly",
    SchedulerMode.Host: "cutlass::gemm::kernel::GroupScheduleMode::kHostPrecompute",
}


ShortSchedulerModeNames = {SchedulerMode.Device: "Device", SchedulerMode.Host: "Host"}


ConvKindTag = {
    cutlass_bindings.conv.Operator.fprop: "cutlass::conv::Operator::kFprop",
    cutlass_bindings.conv.Operator.dgrad: "cutlass::conv::Operator::kDgrad",
    cutlass_bindings.conv.Operator.wgrad: "cutlass::conv::Operator::kWgrad",
}


ConvKindNames = {
    cutlass_bindings.conv.Operator.fprop: "fprop",
    cutlass_bindings.conv.Operator.dgrad: "dgrad",
    cutlass_bindings.conv.Operator.wgrad: "wgrad",
}


IteratorAlgorithmTag = {
    cutlass_bindings.conv.IteratorAlgorithm.analytic: "cutlass::conv::IteratorAlgorithm::kAnalytic",
    cutlass_bindings.conv.IteratorAlgorithm.optimized: "cutlass::conv::IteratorAlgorithm::kOptimized",
    cutlass_bindings.conv.IteratorAlgorithm.fixed_channels: "cutlass::conv::IteratorAlgorithm::kFixedChannels",
    cutlass_bindings.conv.IteratorAlgorithm.few_channels: "cutlass::conv::IteratorAlgorithm::kFewChannels",
}


IteratorAlgorithmNames = {
    cutlass_bindings.conv.IteratorAlgorithm.analytic: "analytic",
    cutlass_bindings.conv.IteratorAlgorithm.optimized: "optimized",
    cutlass_bindings.conv.IteratorAlgorithm.fixed_channels: "fixed_channels",
    cutlass_bindings.conv.IteratorAlgorithm.few_channels: "few_channels",
}


class StrideSupport(enum.Enum):
    Strided = enum_auto()
    Unity = enum_auto()


StrideSupportTag = {
    StrideSupport.Strided: "cutlass::conv::StrideSupport::kStrided",
    StrideSupport.Unity: "cutlass::conv::StrideSupport::kUnity",
}


StrideSupportNames = {
    StrideSupport.Strided: "",
    StrideSupport.Unity: "unity_stride",
}


class ConvMode(enum.Enum):
    CrossCorrelation = enum_auto()
    Convolution = enum_auto()


ConvModeTag = {
    ConvMode.CrossCorrelation: "cutlass::conv::Mode::kCrossCorrelation",
    ConvMode.Convolution: "cutlass::conv::Mode::kConvolution",
}


class MathInstruction:
    """
    Description of a the lowest-level matrix-multiply-accumulate operation to be used in a kernel
    """

    def __init__(
        self,
        instruction_shape,
        element_a,
        element_b,
        element_accumulator,
        opcode_class=cutlass_bindings.OpClass.Simt,
        math_operation=MathOperation.multiply_add,
    ):
        """
        :param instruction_shape: size of the [M, N, K] dimensions of the instruction
        :type instruction_shape: list or tuple
        :param element_a: data type of operand A
        :param element_b: data type of operand B
        :param element_accumulator: data type used in accumulation
        :param opcode_class: higher-level class of the instruction (e.g., SIMT or Tensor Core)
        :type opcode_class: cutlass_bindings.OpClass
        :param math_operation: the type of low-level operation to be performed (e.g., multiply accumulate)
        :type math_operation: MathOperation
        """
        self.instruction_shape = instruction_shape
        self.element_a = element_a
        self.element_b = element_b
        self.element_accumulator = element_accumulator
        self.opcode_class = opcode_class
        self.math_operation = math_operation


class TileDescription:
    """
    Description of a tile of computation to be performed in the kernel, encompassing threadblock, cluster, and warp shapes,
    stage count, and math instruction specification
    """

    def __init__(
        self,
        threadblock_shape,
        stages,
        warp_count,
        math_instruction,
        cluster_shape=[1, 1, 1],
        kernel_schedule: KernelScheduleType = None
    ):
        """
        :param threadblock_shape: shape of a threadblock tyle
        :type threadblock_shape: list or tuple
        :param stages: number of pipline stages in the operation. For SM90 kernels, this can be set to `None` and the maximum
                       number of stages that can be supported for an operation on a given architecture will be computed at a later time
        :type stages: int or None
        :param warp_count: number of warps in each [M, N, K] dimension of a threadblock tile
        :type warp_count: list, tuple, or None
        :param math_instruction: specification of the instruction type and shape to be performed and the types of its operands
        :type math_instruction: MathInstruction
        :param cluster_shape: number of threadblocks in the [X, Y, Z] dimensions of a threadblock cluster
        :param kernel_schedule: type of kernel schedule to use (only available for SM90+)
        :type kernel_schedule: cutlass.backend.KernelScheduleType
        """
        self.threadblock_shape = threadblock_shape
        self.cluster_shape = cluster_shape
        self.kernel_schedule = kernel_schedule
        self.stages: int = stages

        self.math_instruction = math_instruction

        # Number of warps along x, y, z directions
        self.warp_count = warp_count

    @property
    def num_threads(self):
        """
        Returns the number of threads in the threadblock

        :return: number of threads in the threadblock
        :rtype: int or None (if warp count is None)
        """
        if self.warp_count is not None:
            threads = 32
            for cnt in self.warp_count:
                threads *= cnt
            return threads
        return None

    def procedural_name(self):
        """
        Returns a name identifying the tile description

        :return: name identifying the tile description
        :rtype: int
        """
        emit_stages = 0 if self.stages is None else self.stages
        name = "%dx%dx%d_%dx%d_%dx%d" % (
            self.cluster_shape[0],
            self.cluster_shape[1],
            self.cluster_shape[2],
            self.threadblock_shape[0],
            self.threadblock_shape[1],
            self.threadblock_shape[2],
            emit_stages
        )

        return name

    def __str__(self):
        """
        Returns a string with containing each of the tile description's values

        :return: contents of tile description
        :rtype: str
        """
        schedule = KernelScheduleType.ScheduleAuto
        if self.kernel_schedule is not None:
            schedule = self.kernel_schedule
        return f"""
{{
  ClusterShape: {self.cluster_shape}
  ThreadblockShape: {self.threadblock_shape}
  WarpCount: {self.warp_count}
  Stages: {self.stages if self.stages is not None else 'Auto'}
  Kernel schedule: {schedule.name}
}}"""


class TensorDescription:
    def __init__(self, element, layout, alignment=1,
                 complex_transform=cutlass_bindings.complex_transform.none):
        self.element = element
        self.layout = layout
        self.alignment = min(128 // DataTypeSize[self.element], alignment)
        self.complex_transform = complex_transform


def CalculateSmemUsagePerStage(operation):
    """
    Returns the amount of shared memory in bytes consumed in a single stage of a kernel.

    :param op: operation for which the maximum stages should be computed. If stages are
               set via the `op.tile_description.stages` parameter, this setting is ignored
               in the present calculation
    :type op: cutlass.backend.Operation

    :return: number of bytes of shared memory consumed by a single stage
    :rtype: int
    """
    m, n, k = operation.tile_description.threadblock_shape

    if operation.operation_kind == OperationKind.Gemm:
        stage_barrier_bytes = 32
        return (
            (DataTypeSize[operation.A.element] * m * k // 8)
            + (DataTypeSize[operation.B.element] * k * n // 8)
            + stage_barrier_bytes
        )
    else:
        raise Exception("Unsupported operation kind {}.".format(operation.operation_kind))


def CalculateSmemUsage(operation):
    """
    Returns the amount of shared memory in bytes consumed by a kernel.

    :param op: operation for which the maximum stages should be computed. If stages are
               set via the `op.tile_description.stages` parameter, this setting is ignored
               in the present calculation
    :type op: cutlass.backend.Operation

    :return: int
    """
    return operation.tile_description.stages * CalculateSmemUsagePerStage(operation)


class ApiVersion(enum.Enum):
    """
    Differentiate between CUTLASS 2.x and 3.x API versions
    """

    v2x = enum_auto()
    v3x = enum_auto()


def api_version(arch, opclass, datatype):
    """
    Returns whether the architecture, opcode class, and datatype in question require using CUTLASS 2.x
    or 3.x for code emission.

    :param arch: compute capability of device on which to run
    :type arch: int
    :param opclass: class of the operation being performed
    :type opclass: cutlass_bindings.OpClass
    :param datatype: data type to be used in operation (assumes that ElementA and ElementB are the same)

    :return: API version to be used in code emission
    :rtype: ApiVersion
    """
    if (arch >= 90 and
        opclass == cutlass_bindings.OpClass.TensorOp and
        (datatype != cutlass_bindings.float64)):
        return ApiVersion.v3x
    else:
        return ApiVersion.v2x
